"""
MLine (:mod:`skrf.media.MLine`)
========================================

.. autosummary::
   :toctree: generated/

   MLine

"""
from __future__ import annotations

import warnings
from typing import TYPE_CHECKING

import numpy as np
from numpy import arctan, cosh, exp, imag, log, log10, ones, real, sqrt, tanh, zeros
from scipy.constants import c, epsilon_0, mu_0, pi

from ..constants import NumberLike
from ..tlineFunctions import skin_depth, surface_resistivity
from .media import Media

if TYPE_CHECKING:
    from ..frequency import Frequency


class MLine(Media):
    r"""
    A microstripline transmission line defined in terms of width, thickness
    and height on a given relative permittivity substrate. The line has a
    conductor resistivity and a tangential loss factor.

    This class is highly inspired from the technical documentation [#]_
    and sources provided by the qucs project [#]_ .

    In addition, Djordjevic [#]_ /Svensson [#]_  wideband debye dielectric
    model is considered to provide more realistic modelling of broadband
    microstrip with as causal time domain response.

    A compatibility mode is provided to mimic the behaviour of QUCS or of
    Keysight ADS. There is known differences in the output of these
    simulators.

    The quasi-static models of chercteristic impedance and effective
    permittivity give the value at zero frequency. The dispersion models
    compute a frequency-dependant values of these variables.

    * Quasi-static characteristic impedance and effective permittivity models:

        + Hammerstad and Jensen (default)
        + Schneider
        + Wheeler

    * Frequency dispersion of impedance and effective permittivity models:

        + Hammerstad and Jensen
        + Kirschning and Jansen (default)
        + Kobayashi
        + Schneider
        + Yamashita
        + (No dispersion)

    * Strip thickness correction model:

        + all quasi-static models add a certain amount to W to accound for
          non-zero thickness of the strip. Computation with zero  thickness is
          possible.

    Parameters
    ----------
    frequency : :class:`~skrf.frequency.Frequency` object
        frequency band of the media
    z0_port : number, array-like, or None
        `z0_port` is the port impedance for networks generated by the media.
        If `z0_port` is not None, the networks generated by the media are
        renormalized (or in other words embedded) from the characteristic
        impedance z0 of the media to `z0_port`.
        Else if `z0_port` is None, the networks port impedances will be the raw
        characteristic impedance z0 of the media.
        (Default is None)
    z0_override : number, array-like, or None
        `z0_override` override the characteristic impedance for the media.
        If `z0_override` is not None, the networks generated by the media have
        their characteristic impedance `z0` overrided by `z0_override`.
        (Default is None)
    z0 : number, array-like, or None
        deprecated parameter, alias to `z0_override` if `z0_override` is None.
        Emmit a deprecation warning.
    w : number, or array-like
        width of conductor, in m
    h : number, or array-like
        height of substrate between ground plane and conductor, in m
    t : number, or array-like or None, optional
        conductor thickness, in m. Default is None (no width correction
        to account strip thickness).
    ep_r : number, or array-like
        relative permittivity of dielectric at frequency f_epr_tand, no unit
    mu_r : number, array-like
        relative permeability mof dielectric, no unit
    model : str
        microstripline quasi-static impedance and dielectric model in:

        * 'hammerstadjensen' (default)
        * 'schneider'
        * 'wheeler'

    disp : str
        microstripline impedance and dielectric frequency dispersion model in:

        * 'hammerstadjensen'
        * 'kirschningjansen' (default)
        * 'kobayashi'
        * 'schneider'
        * 'yamashita'
        * 'none'

    diel : str
        dielectric frequency dispersion model in:

        * 'djordjevicsvensson' (default)
        * 'frequencyinvariant'

    rho: number, or array-like, optional
        resistivity of conductor, ohm / m
    tand : number, or array-like
        dielectric loss factor at frequency f_epr_tand
    rough : number, or array-like
        RMS roughness of conductor in m
    f_low : number, or array-like
        lower frequency for wideband Debye Djordjevic/Svensson dielectric
        model, in Hz
    f_high : number, or array-like
        higher frequency for wideband Debye Djordjevic/Svensson dielectric
        model, in Hz
    f_epr_tand : number, or array-like
        measurement frequency for ep_r and tand of dielectric, in Hz
    compatibility_mode: str or None (default)
        If set to 'qucs', following behavious happens :

        * Characteristic impedance will be real (no imaginary part due to tand)
        * Quasi-static relative permittivity and impedance will by used for
          loss computation instead of frequency-dispersed values
        *  Kobayashi and Yamashita models will disperse permittivity but keep
           quasi-static impedance values

    \*args, \*\*kwargs : arguments, keyword arguments
            passed to :class:`~skrf.media.media.Media`'s constructor
            (:func:`~skrf.media.media.Media.__init__`

    Note
    ----
    In the case dispersion model only include effective permittivity, no
    dispersion is used for impedance in QUCS mode and Kirschning Jansen is
    used in ADS mode. QUCS mode is the default.

    When the thickness of the strip is smaller than 3 skin depth, the losses
    model gives over-optimistic results and the media will issue a warning.
    At DC, the losses of the line could be smaller than its conductor
    resistance, which is not physical.

    References
    ----------
    .. [#] http://qucs.sourceforge.net/docs/technical.pdf
    .. [#] https://github.com/Qucs/qucsator/blob/develop/src/components/microstrip/msline.cpp
    .. [#] E. Hammerstad and Ø. Jensen,
        "Accurate Models for Microstrip Computer-Aided Design",
        Symposium on Microwave Theory and Techniques, pp. 407-409, June 1980.
    .. [#] M. Kirschning and R. H. Jansen,
        "Accurate Model for Effective Dielectric Constant of Microstrip with
        Validity up to Millimeter-Wave Frequencies", Electronics Letters,
        vol. 8, no. 6, pp. 272-273, Mar. 1982.
    .. [#] R. H. Jansen and M. Kirschning,
        "Arguments and an accurate Model for the Power-Current Formulation of
        Microstrip Characteristic Impedance",
        Archiv für Elektronik und Übertragungstechnik (AEÜ), vol. 37,
        pp. 108-112, 1983.
    .. [#] M. Kobayashi,
        "A Dispersion Formula Satisfying Recent Requirements in Microstrip
        CAD", IEEE Trans. on Microwave Theory and Techniques, vol. 36, no. 8,
        pp. 1246-1250, Aug. 1988.
    .. [#] M. V. Schneider,
        "Microstrip Lines for Microwave Integrated Circuits",
        The Bell System Technical Journal, vol. 48, pp. 1421-1444, May 1969.
    .. [#] M. V. Schneider, "Microstrip Dispersion", Proceedings of the IEEE,
        Letters, vol. 60, Jan. 1972, pp. 144-146.
    .. [#] H. A. Wheeler,
        "Transmission-Line Properties of a Strip on a Dielectric Sheet on a
        Plane, IEEE Trans. on Microwave Theory and Techniques, vol. 25, no. 8,
        pp. 631-647, Aug. 1977.
    .. [#] H. A. Wheeler, "Formulas for the Skin Effect,"
        Proceedings of the IRE, vol. 30, no. 9, pp. 412-424, Sept. 1942.
    .. [#] E. Yamashita, K. Atsuki, and T. Ueda,
        "An Approximate Dispersion Formula of Microstrip Lines for Computer
        Aided Design of Microwave Integrated Circuits", IEEE Trans. on
        Microwave Theory and Techniques, vol. 27, pp. 1036-1038, Dec. 1979.
    .. [#] C. Svensson, G.E. Dermer,
        Time domain modeling of lossy interconnects,
        IEEE Trans. on Advanced Packaging, May 2001, N2, Vol. 24, pp.191-196.
    .. [#] Djordjevic, R.M. Biljic, V.D. Likar-Smiljanic, T.K. Sarkar,
        Wideband frequency-domain characterization of FR-4 and time-domain
        causality,
        IEEE Trans. on EMC, vol. 43, N4, 2001, p. 662-667.
    """
    def __init__(self, frequency: Frequency | None = None,
                 z0_port: NumberLike | None = None,
                 z0_override: NumberLike | None = None,
                 z0: NumberLike | None = None,
                 w: NumberLike = 3, h: NumberLike = 1.6,
                 t: NumberLike | None = None,
                 ep_r: NumberLike = 4.5,
                 mu_r: NumberLike = 1.0,
                 model: str = 'hammerstadjensen',
                 disp: str = 'kirschningjansen',
                 diel: str = 'djordjevicsvensson',
                 rho: NumberLike = 1.68e-8, tand: NumberLike = 0,
                 rough: NumberLike = 0.15e-6,
                 f_low: NumberLike = 1e3, f_high: NumberLike = 1e12,
                 f_epr_tand: NumberLike = 1e9,
                 compatibility_mode: str | None = None,
                 *args, **kwargs):
        Media.__init__(self, frequency = frequency,
                       z0_port = z0_port, z0_override = z0_override, z0 = z0)

        self.w, self.h, self.t = w, h, t
        self.ep_r, self.mu_r  = ep_r, mu_r
        self.model, self.disp, self.diel = model, disp, diel
        self.rho, self.tand, self.rough, self.disp =  rho, tand, rough, disp
        self.f_low, self.f_high, self.f_epr_tand = f_low, f_high, f_epr_tand
        self.compatibility_mode = compatibility_mode

        # variation ofeffective permittivity with frequency
        # Not implemented on QUCS but implemented on ADS.
        # 'frequencyinvariant' will give a constant complex value whith a real
        # part compatible with qucs and an imaginary part due to tand
        self.ep_r_f, self.tand_f = self.analyse_dielectric(
            self.ep_r, self.tand,
            self.f_low, self.f_high, self.f_epr_tand, self.frequency.f,
            self.diel)

        # quasi-static effective permittivity of substrate + line and
        # the impedance of the microstrip line
        # qucs use real-valued ep_r giving real-valued impedance
        if compatibility_mode == 'qucs':
            self.zl_eff, self.ep_reff, self.w_eff = self.analyse_quasi_static(
                real(self.ep_r_f), self.w, self.h, self.t, self.model)
        # ads use complex permittivity giving complex impedance and
        # effective permittivity
        else:
            self.zl_eff, self.ep_reff, self.w_eff = self.analyse_quasi_static(
                self.ep_r_f, self.w, self.h, self.t, self.model)

        # analyse dispersion of impedance and relatice permittivity
        # qucs use w here, but w_eff seems better
        if compatibility_mode == 'qucs':
            self._z_characteristic, self.ep_reff_f = self.analyse_dispersion(
                self.zl_eff, self.ep_reff, real(self.ep_r_f),
                self.w, self.w_eff, self.h, self.t,
                self.frequency.f, self.disp)
        else:
            self._z_characteristic, self.ep_reff_f = self.analyse_dispersion(
                self.zl_eff, self.ep_reff, self.ep_r_f,
                self.w_eff, self.w_eff, self.h, self.t,
                self.frequency.f, self.disp)

        # analyse losses of line
        # qucs use quasi-static values here, leading to a difference
        # against ads
        if compatibility_mode == 'qucs':
            self.alpha_conductor, self.alpha_dielectric = self.analyse_loss(
                real(self.ep_r_f), real(self.ep_reff), self.tand_f,
                self.rho, self.mu_r,
                real(self.zl_eff), real(self.zl_eff),
                self.frequency.f, self.w, self.t, self.rough)
        else:
            self.alpha_conductor, self.alpha_dielectric = self.analyse_loss(
                real(self.ep_r_f), real(self.ep_reff_f), self.tand_f,
                self.rho, self.mu_r,
                real(self._z_characteristic), real(self._z_characteristic),
                self.frequency.f, self.w, self.t, self.rough)

    def __str__(self) -> str:
        f=self.frequency
        output = (
                f'Microstripline Media.  {f.f_scaled[0]}-{f.f_scaled[-1]} {f.unit}.  {f.npoints} points'
                f'\n W= {self.w:.2e}m, H= {self.h:.2e}m')
        return output

    def __repr__(self) -> str:
        return self.__str__()


    @property
    def gamma(self):
        """
        Propagation constant.

        Returns
        -------
        gamma : :class:`numpy.ndarray`
        """
        ep_reff, f = real(self.ep_reff_f), self.frequency.f

        alpha = self.alpha_dielectric.copy()
        if self.rho is not None:
            alpha += self.alpha_conductor

        beta  = 2 * pi * f* sqrt(ep_reff) / c

        return alpha + 1j*beta

    @property
    def z0_characteristic(self) -> np.ndarray:
        """
        Characteristic Impedance, :math:`z_0`

        Returns
        -------
        z0_characteristic : np.ndarray
            Characteristic Impedance in units of ohms
        """
        if self.z0_override is None:
            return self._z_characteristic
        else:
            return self.z0_override

    @property
    def Z0_f(self) -> np.ndarray:
        """
        Alias fos Characteristic Impedance for backward compatibility.
        Deprecated, do not use.

        Returns
        -------
        z0 : :class:`numpy.ndarray`
        """
        warnings.warn(
            "`Z0_f` is deprecated, use `Z0` instead",
             DeprecationWarning, stacklevel = 2
        )
        return self._z_characteristic

    def analyse_dielectric(self, ep_r: NumberLike, tand: NumberLike,
                          f_low: NumberLike, f_high: NumberLike,
                          f_epr_tand: NumberLike, f: NumberLike,
                          diel: str):
        """
        This function calculate the frequency dependent relative permittivity
        of dielectric and and tangeantial loss factor.

        References
        ----------
        .. [#] C. Svensson, G.E. Dermer,
            Time domain modeling of lossy interconnects,
            IEEE Trans. on Advanced Packaging, May 2001, N2, Vol. 24, pp.191-196.
        .. [#] Djordjevic, R.M. Biljic, V.D. Likar-Smiljanic, T.K. Sarkar,
            Wideband frequency-domain characterization of FR-4 and time-domain
            causality,
            IEEE Trans. on EMC, vol. 43, N4, 2001, p. 662-667.

        Returns
        -------
        ep_r_f : :class:`numpy.ndarray`
        tand_f : :class:`numpy.ndarray`
        """
        if diel == 'djordjevicsvensson':
            # compute the slope for a log frequency scale, tanD dependent.
            k = log((f_high + 1j * f_epr_tand) / (f_low + 1j * f_epr_tand))
            fd = log((f_high + 1j * f) / (f_low + 1j * f))
            ep_d = -tand * ep_r  / imag(k)
            # value for frequency above f_high
            ep_inf = ep_r * (1. + tand * real(k) / imag(k))
            # compute complex permitivity
            ep_r_f = ep_inf + ep_d * fd
            # get tand
            tand_f = -imag(ep_r_f) / real(ep_r_f)
        elif diel == 'frequencyinvariant':
            ep_r_f =  ep_r - 1j * ep_r * tand
            tand_f = tand
        else:
            raise ValueError('Unknown dielectric dispersion model')

        return ep_r_f, tand_f

    def analyse_quasi_static(self, ep_r: NumberLike,
                           w: NumberLike, h: NumberLike, t: NumberLike,
                           model: str):
        """
        This function calculates the quasi-static impedance of a microstrip
        line, the value of the effective permittivity as per filling factor
        and the effective width due to the finite conductor thickness for the
        given microstrip line and substrate properties.

        References
        ----------
        .. [#] E. Hammerstad and Ø. Jensen,
            "Accurate Models for Microstrip Computer-Aided Design", Symposium
            on Microwave Theory and Techniques, pp. 407-409, June 1980.
        .. [#] H. A. Wheeler,
            "Transmission-Line Properties of a Strip on a Dielectric Sheet on a
            Plane, IEEE Trans. on Microwave Theory and Techniques, vol. 25,
            no. 8, pp. 631-647, Aug. 1977.
        .. [#] M. V. Schneider,
            "Microstrip Lines for Microwave Integrated Circuits",
            The Bell System Technical Journal, vol. 48, pp. 1421-1444, May 1969.

        Returns
        -------
        zl_eff : :class:`numpy.ndarray`
        ep_reff : :class:`numpy.ndarray`
        """
        Z0 = sqrt(mu_0 / epsilon_0)
        zl_eff = Z0
        ep_reff = ep_r
        w_eff = w

        if model == 'wheeler':
            # compute strip thickness effect
            dw1 = 0
            if t is not None and t > 0:
                dw1 = t / pi * log(4. * exp(1.) / sqrt((t / h)**2) + \
                                   (1. / pi / (w / t + 1.1))**2)
            dwr = (1. + 1. / ep_r) / 2. * dw1
            wr = w + dwr
            w_eff = wr

            # compute characteristic impedance
            if (w / h) < 3.3:
                cp = log(4. * h / wr + sqrt((4 * h / wr)**2 + 2))
                b = (ep_r - 1.) / (ep_r + 1.) / \
                    2 * (log(pi / 2.) + log(4. / pi) / ep_r)
                zl_eff = (cp - b) * Z0 / pi / sqrt(2 * (ep_r + 1.))
            else:
                cp = 1 + log(pi / 2.) + log(wr / h / 2. + 0.94)
                d = 1. / pi / 2. * (1. + log(pi**2 / 16.)) * (ep_r - 1.) \
                    / ep_r**2
                x = 2. * log(2.) / pi + wr / h / 2. + (ep_r + 1.) / 2 / pi / \
                ep_r * cp + d
                zl_eff = Z0 / 2 / x / sqrt(ep_r)

            # compute effective dielectric constant
            if (w / h) < 1.3:
                a = log(8 * h / wr) + (wr / h)**2 / 32
                b = (ep_r - 1.) / (ep_r + 1.) / \
                    2 * (log(pi / 2.) + log(4. / pi) / ep_r)
                ep_reff = (ep_r + 1.) / 2. * (a / (a - b))**2
            else:
                # qucsator is 4.0137 but doc 0.94 * 2 = 1.88
                d = (ep_r - 1.) / 2. / pi / ep_r * \
                    (log(2.1349 * wr / h + 4.0137) - 0.5169 / ep_r)
                e = wr / h / 2 + 1. / pi * log(8.5397 * wr / h + 16.0547)
                ep_reff = ep_r * ((e - d) / e)**2

        elif model == 'schneider':
            u = w / h
            dw = 0

            # consider strip thickness equations
            if t is not None and t > 0:
                if t < (w / 2):
                    if u < (1. / pi / 2):
                        arg = 2 * pi * w / t
                    else:
                        arg = h / t
                    dw = t / pi * (1. + log(2 * arg))
                    if (t / dw) >= 0.75:
                        dw = 0
            w_eff = w + dw
            u = w_eff / h

            # effective dielectric constant
            ep_reff = (ep_r + 1.) / 2. + (ep_r - 1.) / 2. / sqrt (1. + 10. / u)

            # characteristic impedance
            if u < 1.:
                z = 1. / pi / 2. * log(8. / u + u / 4)
            else:
                z = 1. / (u + 2.42 - 0.44 / u + (1. - 1. / u)**6)
            zl_eff = Z0 * z / sqrt(ep_reff)

        elif model == 'hammerstadjensen':
            u = w / h
            if t is not None:
                t = t/h
            du1 = 0.

            # compute strip thickness effect
            if t is not None and t > 0:
                # Qucs formula 11.22 is wrong, normalized w has to be used instead (see Hammerstad and Jensen Article)
                # Normalized w is named u and is actually used in qucsator source code
                # coth(alpha) = 1/tanh(alpha)
                du1 = t / pi * log(1. + 4. * exp(1.) / t * tanh(sqrt(6.517 * u))**2)

            # sech(alpha) = 1/cosh(alpha)
            dur =  du1 * (1. + 1. / cosh(sqrt(ep_r - 1.))) / 2.

            u1 = u + du1
            ur = u + dur
            w_eff = ur * h

            # compute impedances for homogeneous medium
            zr = hammerstad_zl(ur)
            z1 = hammerstad_zl(u1)

            # compute effective dielectric constant
            a, b  = hammerstad_ab(ur, ep_r)
            e = hammerstad_er(ur, ep_r, a, b)

            # compute final characteristic impedance and dielectric constant
            #including strip thickness effects
            zl_eff = zr / sqrt(e)
            ep_reff = e * (z1 / zr)**2

        else:
            raise ValueError('Unknown microstripline quasi-static model')

        return zl_eff, ep_reff, w_eff

    def analyse_dispersion(self, zl_eff: NumberLike, ep_reff: NumberLike,
                          ep_r: NumberLike, wr: NumberLike, w_eff: NumberLike,
                          h: NumberLike, t: NumberLike, f: NumberLike,
                          disp: str):
         """
         This function compute the frequency dependent characteristic
         impedance and effective permittivity accounting for microstripline
         frequency dispersion.

         References
         ----------
         .. [#] M. Kobayashi,
             "A Dispersion Formula Satisfying Recent Requirements in Microstrip
             CAD", IEEE Trans. on Microwave Theory and Techniques, vol. 36,
             no. 8, pp. 1246-1250, Aug. 1988.
         .. [#] M. V. Schneider, "Microstrip Dispersion", Proceedings of the
             IEEE, Letters, vol. 60, Jan. 1972, pp. 144-146.
         .. [#] M. Kirschning and R. H. Jansen,
             "Accurate Model for Effective Dielectric Constant of Microstrip
             with Validity up to Millimeter-Wave Frequencies", Electronics
             Letters, vol. 8, no. 6, pp. 272-273, Mar. 1982.
         .. [#] R. H. Jansen and M. Kirschning,
             "Arguments and an accurate Model for the Power-Current Formulation of
             Microstrip Characteristic Impedance",
             Archiv für Elektronik und Übertragungstechnik (AEÜ), vol. 37,
             pp. 108-112, 1983.
         .. [#] E. Yamashita, K. Atsuki, and T. Ueda,
             "An Approximate Dispersion Formula of Microstrip Lines for
             Computer Aided Design of Microwave Integrated Circuits",
             IEEE Trans. on Microwave Theory and Techniques, vol. 27,
             pp. 1036-1038, Dec. 1979.

         Returns
         -------
         z : :class:`numpy.ndarray`
         e : :class:`numpy.ndarray`
         """
         u = wr/h
         if disp == 'schneider':
             k = sqrt(ep_reff / ep_r)
             fn = 4. * h * f / c * sqrt(ep_r - 1.)
             fn2 = fn**2
             e = ep_reff * ((1. + fn2) / (1. + k * fn2))**2
             z = zl_eff * sqrt(ep_reff / e)
         elif disp == 'hammerstadjensen':
             Z0 = sqrt(mu_0 / epsilon_0)
             g = pi**2 / 12 * (ep_r - 1) / ep_reff * sqrt(2 * pi * zl_eff / Z0)
             fp = (2 * mu_0 * h * f) / zl_eff
             e = ep_r - (ep_r - ep_reff) / (1 + g * fp**2)
             z =  zl_eff * sqrt(ep_reff / e) * (e - 1) / (ep_reff - 1)
         elif disp == 'kirschningjansen':
             fn = f * h * 1e-6
             e = kirsching_er(u, fn, ep_r, ep_reff)
             z, _ = kirsching_zl(u, fn, ep_r, ep_reff, e, zl_eff)
         elif disp == 'yamashita':
             k = sqrt(ep_r / ep_reff)
             fp = 4 * h * f / c * sqrt(ep_r - 1) * \
                 (0.5 + (1 + 2 * log10(1 + u))**2)
             e = ep_reff * ((1 + k * fp**1.5 / 4) / (1 + fp**1.5 / 4))**2
             # qucs keep quasi-static impedance here
             if self.compatibility_mode == 'qucs':
                 z =  np.ones(f.shape) * zl_eff
             # use Kirschning Jansen for impedance dispersion by default
             else:
                 fn = f * h * 1e-6
                 z, _ = kirsching_zl(wr / h, fn, ep_r, ep_reff, e, zl_eff)
         elif disp == 'kobayashi':
             fk = c * arctan(ep_r * sqrt((ep_reff - 1) / (ep_r - ep_reff)))/ \
                 (2 * pi * h * sqrt(ep_r - ep_reff))
             fh = fk / (0.75 + (0.75 - 0.332 / (ep_r**1.73)) * u)
             no = 1 + 1 / (1 + sqrt(u)) + 0.32 * (1 / (1 + sqrt(u)))**3
             nc = np.where(u < 0.7,
                 1 + 1.4 / (1 + u) * (0.15 - 0.235 * exp(-0.45 * f / fh)),
                 1)
             n = np.where(no * nc < 2.32, no * nc, 2.32)
             e =  ep_r - (ep_r - ep_reff) / (1 + (f / fh)**n)
             # qucs keep quasi-static impedance here
             if self.compatibility_mode == 'qucs':
                 z =  np.ones(f.shape) * zl_eff
             # use Kirschning Jansen for impedance dispersion by default
             else:
                 fn = f * h * 1e-6
                 z, _ = kirsching_zl(wr / h, fn, ep_r, ep_reff, e, zl_eff)
         elif disp == 'none':
             e = ones(f.shape) * ep_reff
             z = ones(f.shape) * zl_eff

         else:
             raise ValueError('Unknown microstripline dispersion model')

         return z, e

    def analyse_loss(self, ep_r: NumberLike, ep_reff: NumberLike,
                    tand: NumberLike, rho: NumberLike, mu_r: NumberLike,
                    zl_eff_f1: NumberLike, zl_eff_f2: NumberLike,
                    f: NumberLike, w: NumberLike, t: NumberLike,
                    D: NumberLike):
        """
        The function calculates the conductor and dielectric losses of a
        single microstrip line using wheeler's incremental inductance rule.

        References
        ----------
        .. [#] H. A. Wheeler, "Formulas for the Skin Effect,"
            Proceedings of the IRE, vol. 30, no. 9, pp. 412-424, Sept. 1942.

        Returns
        -------
        a_conductor : :class:`numpy.ndarray`
        a_dielectric : :class:`numpy.ndarray`
        """
        # limited to only Hammerstad and Jensen model
        Z0 = np.sqrt(mu_0/epsilon_0)

        # conductor losses
        if t is not None and  t > 0:
            if rho is None:
                raise(AttributeError('must provide resistivity rho. '
                                     'see initializer help'))
            else:
                Rs  = surface_resistivity(f=f, rho=rho, mu_r=1)
                ds = skin_depth(f, rho, mu_r)
                if(np.any(t < 3 * ds)):
                    warnings.warn(
                        'Conductor loss calculation invalid for line'
                        f'height t ({t})  < 3 * skin depth ({ds[0]})',
                        RuntimeWarning, stacklevel=2
                        )
                # current distribution factor
                Ki  = exp(-1.2 * ((zl_eff_f1 + zl_eff_f2) / 2 / Z0)**0.7)
                # D is RMS surface roughness
                Kr  = 1 + 2 / pi * arctan(1.4 * (D/ds)**2)
            a_conductor = Rs / (zl_eff_f1 * w) * Ki * Kr
        else:
            a_conductor = zeros(f.shape)

        # dielectric losses
        l0 = c / f
        a_dielectric =  pi * ep_r / (ep_r - 1) * (ep_reff - 1) / \
            sqrt(ep_reff) * tand / l0

        return a_conductor, a_dielectric

def hammerstad_ab(u: NumberLike, ep_r: NumberLike) -> NumberLike:
    """
    Hammerstad parameters for relative permittivity dispersion.
    """
    a = 1. + log((u**4 + (u / 52.)**2) / (u**4 + 0.432)) / 49. \
        + log(1 + (u / 18.1)**3) / 18.7

    b = 0.564 * ((ep_r - 0.9) / (ep_r + 3.))**0.053

    return a, b

def hammerstad_zl(u: NumberLike) -> NumberLike:
    """
    Hammerstad quasi-static impedance.
    """
    fu = 6 + (2 * pi - 6) * exp(-(30.666 / u)**0.7528)
    Z0 = sqrt(mu_0/epsilon_0)
    return Z0 / 2. / pi * log(fu / u + sqrt(1. + (2. / u)**2))

def hammerstad_er(u: NumberLike, ep_r: NumberLike, a: NumberLike,
                  b: NumberLike) -> NumberLike:
    """
    Hammerstad quasi-static relative permittivity.
    """
    return (ep_r + 1) / 2 + (ep_r - 1) / 2 * (1. + 10. / u)**(-a * b)

def kirsching_zl(u: NumberLike, fn: NumberLike,
                 ep_r: NumberLike, ep_reff: NumberLike, ep_reff_f: NumberLike,
                 zl_eff: NumberLike):
    """
    Kirschning Jansen impedance dispersion.
    """
    #fn = f * h * 1e-6 # GHz-mm
    R1 = np.minimum(0.03891 * ep_r**1.4, 20.)
    R2 = np.minimum(0.2671 * u**7, 20.)
    R3 = 4.766 * exp(-3.228 * u**0.641)
    R4 = 0.016 + (0.0514 * ep_r)**4.524
    R5 = (fn / 28.843)**12
    R6 = np.minimum(22.20 * u **1.92, 20.)
    R7 = 1.206 - 0.3144 * exp(-R1) * (1 - exp(-R2))
    R8 = 1 + 1.275 * (1 - exp(-0.004625 * R3 * ep_r**1.674 \
                              * (fn / 18.365)**2.745))
    R9 = 5.086 * R4 * R5/(0.3838 + 0.386 * R4) \
        * exp(-R6) / (1 + 1.2992 * R5) \
        * (ep_r - 1)**6 / (1 + 10 * (ep_r - 1)**6)
    R10 = 0.00044 * ep_r**2.136 + 0.0184
    R11 = (fn / 19.47)**6 / (1 + 0.0962 * (fn / 19.47)**6)
    R12 = 1 / (1 + 0.00245 * u**2)
    R13 = 0.9408 * ep_reff_f**R8 - 0.9603
    R14 = (0.9408 - R9) * ep_reff**R8 - 0.9603
    R15 = 0.707 * R10 * (fn / 12.3)**1.097
    R16 = 1 + 0.0503 * ep_r**2 * R11 * (1 - exp(-(u / 15)**6))
    R17 = R7 * (1 - 1.1241 * R12 / R16 \
        *exp(-0.026 * fn**1.15656 - R15))
    return zl_eff * (R13 / R14)**R17, R17

def kirsching_er(u: NumberLike, fn: NumberLike,
                 ep_r: NumberLike, ep_reff: NumberLike):
    """
    Kirschning Jansen relative permittivity dispersion.
    """
    # in the paper fn is in GHz-cm while in Qucs it is GHz-mm, thus a factor
    # 10 for all constant that multiply or divide fn
    P1 = 0.27488 + (0.6315 + 0.525 / ( 1+ 0.0157 * fn)**20) * u \
        -0.065683 * exp(-8.7513 * u)
    P2 = 0.33622 * (1  -exp(-0.03442 * ep_r))
    P3 = 0.0363 * exp(-4.6 * u) * (1 - exp(-(fn / 38.7)**4.97))
    P4 = 1 + 2.751 * (1 - exp(-(ep_r / 15.916)**8))
    Pf = P1 * P2 * ((0.1844 + P3 * P4) * fn)**1.5763
    return ep_r - (ep_r - ep_reff) / (1 + Pf)



import pytest

import skrf as rf

networks = [
    "ntwk1",
    "line",
    "open_2p",
    "short_2p",
    "ind",
    "ring_slot",
    "tee",
    "ring_slot_meas",
    "wr2p2_line" ,
    "wr2p2_line1",
    "wr2p2_delayshort",
    "wr2p2_short",
    "wr1p5_line",
    "wr1p5_short",
    "ro_1",
    "ro_2",
    "ro_3",
]

airs = [
    "air",
    "air50"
]


frequencies = [
    "f_wr51",
    "f_wr42",
    "f_wr34",
    "f_wr28",
    "f_wr22p4",
    "f_wr18p8",
    "f_wr14p8",
    "f_wr12p2",
    "f_wr10",
    "f_wr8",
    "f_wr6p5",
    "f_wr5p1",
    "f_wr4p3",
    "f_wr3p4",
    "f_wr2p8",
    "f_wr2p2",
    "f_wr1p9",
    "f_wr1p5",
    "f_wr1p2",
    "f_wr1",
    "f_wr0p8",
    "f_wr0p65",
    "f_wr0p51",
    "f_wm1295",
    "f_wm1092",
    "f_wm864",
    "f_wm710",
    "f_wm570",
    "f_wm470",
    "f_wm380",
    "f_wm310",
    "f_wm250",
    "f_wm200",
    "f_wm164",
    "f_wm130",
    "f_wm106",
    "f_wm86",
]


waveguides = [
    "wr51",
    "wr42",
    "wr34",
    "wr28",
    "wr22p4",
    "wr18p8",
    "wr14p8",
    "wr12p2",
    "wr10",
    "wr8",
    "wr6p5",
    "wr5p1",
    "wr4p3",
    "wr3p4",
    "wr2p8",
    "wr2p2",
    "wr1p9",
    "wr1p5",
    "wr1p2",
    "wr1",
    "wr0p8",
    "wr0p65",
    "wr0p51",
    "wm1295",
    "wm1092",
    "wm864",
    "wm710",
    "wm570",
    "wm470",
    "wm380",
    "wm310",
    "wm250",
    "wm200",
    "wm164",
    "wm130",
    "wm106",
    "wm86"
]

@pytest.mark.parametrize("name", networks)
def test_static_data(name):
    getattr(rf.data, name)


@pytest.mark.parametrize("name", airs)
def test_static_airs(name):
    getattr(rf, name)

@pytest.mark.parametrize("name", frequencies)
def test_static_frequencies(name):
    getattr(rf, name)

@pytest.mark.parametrize("name", waveguides)
def test_static_waveguides(name):
    getattr(rf, name)



import os
import unittest

import numpy as np
from numpy.testing import assert_array_almost_equal

from skrf.constants import S_DEF_DEFAULT
from skrf.frequency import Frequency
from skrf.media import DefinedGammaZ0
from skrf.network import Network, concat_ports, connect


class DefinedGammaZ0TestCase(unittest.TestCase):
    def setUp(self):
        self.files_dir = os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            'qucs_prj'
            )
        self.dummy_media = DefinedGammaZ0(
            frequency = Frequency(1, 100, 21, unit='ghz'),
            gamma=1j,
            z0 = 50 ,
            )

    def test_impedance_mismatch(self):
        """
        Test the naming of the network. When circuit is used to connect a
        topology of networks, they should have unique names.
        """
        name = 'impedanceMismatch,50to25'
        qucs_ntwk = Network(os.path.join(self.files_dir, name + '.s2p'))
        # sNp-files store the same impedance for every port, so to compare the
        # networks, the port impedance has to be set manually to [50,25]
        qucs_ntwk.z0 = [50,25]
        self.dummy_media.frequency = qucs_ntwk.frequency
        skrf_ntwk = self.dummy_media.thru(z0=50, name = name)**\
            self.dummy_media.thru(z0=25)

        self.assertEqual(qucs_ntwk, skrf_ntwk)
        self.assertEqual(qucs_ntwk.name, skrf_ntwk.name)

    def test_tee(self):
        """
        Test the naming of the network. When circuit is used to connect a
        topology of networks, they should have unique names.
        """
        name = 'tee'
        self.dummy_media.frequency = Frequency(1, 1, 1, unit='GHz')
        skrf_ntwk = self.dummy_media.tee(name = name)
        self.assertEqual(name, skrf_ntwk.name)

    def test_splitter(self):
        """
        Test the naming of the network. When circuit is used to connect a
        topology of networks, they should have unique names.
        Test mismatched splitter.
        """
        name = 'splitter'
        self.dummy_media.frequency = Frequency(1, 1, 1, unit='GHz')
        skrf_ntwk = self.dummy_media.splitter(3, name = name)
        self.assertEqual(name, skrf_ntwk.name)

    def test_mismatch_splitter(self):
        """
        Test mismatched splitter against 50-ohm splitter connected to
        equivalent mismatched thrus.
        """
        freq = Frequency(1, 1, 1, unit='GHz')
        med = DefinedGammaZ0(frequency=freq, z0=50)

        for n in range(2, 5):
            z_port = np.arange(1, n + 1) * 20
            split =med.splitter(n, z0=z_port)
            thrus = []
            for k in range(n):
                thrus.append(med.thru(z0 = z_port[k]))

            # connect
            thru = concat_ports(thrus, port_order='second')
            ref = med.splitter(n)
            ref = connect(ref, 0, thru, 0, n)
            assert_array_almost_equal(split.s, ref.s)

    def test_thru(self):
        """
        Test the naming of the network. When circuit is used to connect a
        topology of networks, they should have unique names.
        """
        name = 'thru'
        self.dummy_media.frequency = Frequency(1, 1, 1, unit='GHz')
        skrf_ntwk = self.dummy_media.thru(name = name)
        self.assertEqual(name, skrf_ntwk.name)

    def test_line(self):
        """
        Test the naming of the network. When circuit is used to connect a
        topology of networks, they should have unique names.
        """
        name = 'line'
        self.dummy_media.frequency = Frequency(1, 1, 1, unit='GHz')
        skrf_ntwk = self.dummy_media.line(90, 'deg', name = name)
        self.assertEqual(name, skrf_ntwk.name)

    def test_delay_load(self):
        """
        Test the naming of the network. When circuit is used to connect a
        topology of networks, they should have unique names.
        """
        name = 'delay_load'
        self.dummy_media.frequency = Frequency(1, 1, 1, unit='GHz')
        skrf_ntwk = self.dummy_media.delay_load(1j, 90, 'deg',
                                                      name = name)
        self.assertEqual(name, skrf_ntwk.name)

    def test_shunt_delay_load(self):
        """
        Test the naming of the network. When circuit is used to connect a
        topology of networks, they should have unique names.
        """
        name = 'shunt_delay_load'
        self.dummy_media.frequency = Frequency(1, 1, 1, unit='GHz')
        skrf_ntwk = self.dummy_media.shunt_delay_load(1j, 90, 'deg',
                                                      name = name)
        self.assertEqual(name, skrf_ntwk.name)

    def test_delay_open(self):
        """
        Test the naming of the network. When circuit is used to connect a
        topology of networks, they should have unique names.
        """
        name = 'delay_open'
        self.dummy_media.frequency = Frequency(1, 1, 1, unit='GHz')
        skrf_ntwk = self.dummy_media.delay_open(90, 'deg', name = name)
        self.assertEqual(name, skrf_ntwk.name)

    def test_shunt_delay_open(self):
        """
        Test the naming of the network. When circuit is used to connect a
        topology of networks, they should have unique names.
        """
        name = 'shunt_delay_open'
        self.dummy_media.frequency = Frequency(1, 1, 1, unit='GHz')
        skrf_ntwk = self.dummy_media.shunt_delay_open(90, 'deg', name = name)
        self.assertEqual(name, skrf_ntwk.name)

    def test_delay_short(self):
        """
        Test the naming of the network. When circuit is used to connect a
        topology of networks, they should have unique names.
        """
        name = 'delay_short'
        self.dummy_media.frequency = Frequency(1, 1, 1, unit='GHz')
        skrf_ntwk = self.dummy_media.delay_short(90, 'deg', name = name)
        self.assertEqual(name, skrf_ntwk.name)

    def test_shunt_delay_short(self):
        """
        Test the naming of the network. When circuit is used to connect a
        topology of networks, they should have unique names.
        """
        name = 'shunt_delay_short'
        self.dummy_media.frequency = Frequency(1, 1, 1, unit='GHz')
        skrf_ntwk = self.dummy_media.shunt_delay_short(90, 'deg', name = name)
        self.assertEqual(name, skrf_ntwk.name)

    def test_resistor(self):
        """
        Compare the component values against s-parameters generated by QUCS.
        Test the naming of the network. When circuit is used to connect a
        topology of networks, they should have unique names.
        """
        name = 'resistor,1ohm'
        qucs_ntwk = Network(os.path.join(self.files_dir, name + '.s2p'))
        self.dummy_media.frequency = qucs_ntwk.frequency
        skrf_ntwk = self.dummy_media.resistor(1, name = name)
        self.assertEqual(qucs_ntwk, skrf_ntwk)
        self.assertEqual(qucs_ntwk.name, skrf_ntwk.name)
        # test vs analytical ABCD parameters of a series resistor
        Z = 10 - 4j
        ntwk = self.dummy_media.resistor(Z)
        ABCD = np.full(ntwk.s.shape, [[1, -1+1j],
                                       [0, 1]])
        ABCD[:,0,1] = Z
        assert_array_almost_equal(ABCD, ntwk.a)

    def test_resistor_mismatch(self):
        """
        Compare the component values against s-parameters generated by QUCS.
        The mismatch QUCS component is generated by renormalize() method.
        """
        # Config the qucs_ntwk and media
        name = 'resistor,1ohm'
        qucs_ntwk = Network(os.path.join(self.files_dir, name + '.s2p'))
        self.dummy_media.frequency = qucs_ntwk.frequency
        mismatch_z0_tuple = ([25, 50], [25-5j, 50+10j], [25, 50-10j])

        # Test against Y-parameter defination
        def resistor_y_def(media: DefinedGammaZ0, R, **kwargs):
            result = media.match(nports=2, **kwargs)
            y = np.zeros(shape=result.s.shape, dtype=complex)
            R = np.array(R)
            y[:, 0, 0] = 1.0 / R
            y[:, 1, 1] = 1.0 / R
            y[:, 0, 1] = -1.0 / R
            y[:, 1, 0] = -1.0 / R
            result.y = y
            return result

        # Test real and imag port impedance with Y-parameter defination and renormalize()
        for mismatch_z0 in mismatch_z0_tuple:
            qucs_ntwk_copy = qucs_ntwk.copy()
            qucs_ntwk_copy.renormalize(mismatch_z0)
            skrf_ntwk = self.dummy_media.resistor(1, name = name, z0=mismatch_z0)
            skrf_ntwk_y = resistor_y_def(self.dummy_media, 1, z0=mismatch_z0)
            self.assertEqual(qucs_ntwk_copy, skrf_ntwk)
            assert_array_almost_equal(skrf_ntwk.s, skrf_ntwk_y.s)

        # Config the s_def parameters
        mismatch_z0 = [25, 50-10j]
        s_defs = ("power", "pseudo", "traveling")

        # Test different s_def configurations
        for s_def in s_defs:
            qucs_ntwk_copy = qucs_ntwk.copy()
            qucs_ntwk_copy.renormalize(mismatch_z0, s_def=s_def)
            skrf_ntwk = self.dummy_media.resistor(1, name=name, z0=mismatch_z0, s_def=s_def)
            skrf_ntwk_y = resistor_y_def(self.dummy_media, 1, z0=mismatch_z0, s_def=s_def)
            self.assertEqual(skrf_ntwk, qucs_ntwk_copy)
            self.assertEqual(skrf_ntwk.s_def, qucs_ntwk_copy.s_def)
            assert_array_almost_equal(skrf_ntwk.s, skrf_ntwk_y.s)
            self.assertEqual(skrf_ntwk.s_def, skrf_ntwk_y.s_def)

    def test_shunt_resistor(self):
        """
        Test the naming of the network. When circuit is used to connect a
        topology of networks, they should have unique names.
        """
        name = 'shunt_resistor,1ohm'
        self.dummy_media.frequency = Frequency(1, 1, 1, unit='GHz')
        skrf_ntwk = self.dummy_media.resistor(1, name = name)
        self.assertEqual(name, skrf_ntwk.name)
        # test vs analytical ABCD parameters of a shunt resistor
        Z = 10 - 4j
        ntwk = self.dummy_media.shunt_resistor(Z)
        ABCD = np.asarray([[[1, 0], [1/Z, 1]]])
        assert_array_almost_equal(ABCD, ntwk.a)

    def test_capacitor(self):
        """
        Compare the component values against s-parameters generated by QUCS.
        Test the naming of the network. When circuit is used to connect a
        topology of networks, they should have unique names.
        """
        name = 'capacitor,p01pF'
        qucs_ntwk = Network(os.path.join(self.files_dir, name + '.s2p'))
        self.dummy_media.frequency = qucs_ntwk.frequency
        skrf_ntwk = self.dummy_media.capacitor(.01e-12, name = name)
        self.assertEqual(qucs_ntwk, skrf_ntwk)
        self.assertEqual(qucs_ntwk.name, skrf_ntwk.name)
        # test vs analytical ABCD parameters of a series capacitor
        C = 0.1e-12
        ntwk = self.dummy_media.capacitor(C)
        Z = 1/(1j*C*ntwk.frequency.w)
        ABCD = np.full(ntwk.s.shape, [[1, -1+1j],
                                       [0, 1]])
        ABCD[:,0,1] = Z
        assert_array_almost_equal(ABCD, ntwk.a)

    def test_capacitor_mismatch(self):
        """
        Compare the component values against s-parameters generated by QUCS.
        The mismatch QUCS component is generated by renormalize() method.
        """
        # Config the qucs_ntwk and media
        name = 'capacitor,p01pF'
        qucs_ntwk = Network(os.path.join(self.files_dir, name + '.s2p'))
        self.dummy_media.frequency = qucs_ntwk.frequency
        mismatch_z0_tuple = ([25, 50], [25-5j, 50+10j], [25, 50-10j])

        # Test against Y-parameter defination
        def capacitor_y_def(media: DefinedGammaZ0, C, **kwargs):
            result = media.match(nports=2, **kwargs)
            w = media.frequency.w
            y = np.zeros(shape=result.s.shape, dtype=complex)
            C = np.array(C)
            y[:, 0, 0] = 1j * w * C
            y[:, 1, 1] = 1j * w * C
            y[:, 0, 1] = -1j * w * C
            y[:, 1, 0] = -1j * w * C
            result.y = y
            return result

        # Test real and imag port impedance with Y-parameter defination and renormalize()
        for mismatch_z0 in mismatch_z0_tuple:
            qucs_ntwk_copy = qucs_ntwk.copy()
            qucs_ntwk_copy.renormalize(mismatch_z0)
            skrf_ntwk = self.dummy_media.capacitor(.01e-12, name = name, z0=mismatch_z0)
            skrf_ntwk_y = capacitor_y_def(self.dummy_media, .01e-12, z0=mismatch_z0)
            self.assertEqual(qucs_ntwk_copy, skrf_ntwk)
            assert_array_almost_equal(skrf_ntwk.s, skrf_ntwk_y.s)

        # Config the s_def parameters
        mismatch_z0 = [25, 50-10j]
        s_defs = ("power", "pseudo", "traveling")

        # Test different s_def configurations
        for s_def in s_defs:
            qucs_ntwk_copy = qucs_ntwk.copy()
            qucs_ntwk_copy.renormalize(mismatch_z0, s_def=s_def)
            skrf_ntwk = self.dummy_media.capacitor(.01e-12, name=name, z0=mismatch_z0, s_def=s_def)
            skrf_ntwk_y = capacitor_y_def(self.dummy_media, .01e-12, z0=mismatch_z0, s_def=s_def)
            self.assertEqual(skrf_ntwk, qucs_ntwk_copy)
            self.assertEqual(skrf_ntwk.s_def, qucs_ntwk_copy.s_def)
            assert_array_almost_equal(skrf_ntwk.s, skrf_ntwk_y.s)
            self.assertEqual(skrf_ntwk.s_def, skrf_ntwk_y.s_def)

    def test_shunt_capacitor(self):
        """
        Test the naming of the network. When circuit is used to connect a
        topology of networks, they should have unique names.
        """
        name = 'shunt_capacitor,p01pF'
        self.dummy_media.frequency = Frequency(1, 1, 1, unit='GHz')
        skrf_ntwk = self.dummy_media.capacitor(.01e-12, name = name)
        self.assertEqual(name, skrf_ntwk.name)
        # test vs analytical ABCD parameters of a shunt capacitor
        C = 0.1e-12
        ntwk = self.dummy_media.shunt_capacitor(C)
        Z = 1/(1j*C*ntwk.frequency.w)
        ABCD = np.full(ntwk.s.shape, [[1, 0],
                                       [-1+1j, 1]])
        ABCD[:,1,0] = 1/Z
        assert_array_almost_equal(ABCD, ntwk.a)

    def test_capacitor_q(self):
        """
        Compare the component values against s-parameters generated by ADS.
        Test the naming of the network. When circuit is used to connect a
        topology of networks, they should have unique names.
        """
        name = 'capacitor_q'
        ads_path = os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            'ads'
            )
        ads_ntwk = Network(os.path.join(ads_path, name + '.s2p'))
        self.dummy_media.frequency = ads_ntwk.frequency
        skrf_ntwk = self.dummy_media.capacitor_q(C=1.0e-12, f_0=1.0e9, q_factor=30.0, name = name)
        self.assertEqual(ads_ntwk, skrf_ntwk)
        self.assertEqual(ads_ntwk.name, skrf_ntwk.name)
        # test vs analytical ABCD parameters of a series capacitor with q_factor
        C = 0.1e-12
        Q = 30.0
        F = 1.0e9
        ntwk = self.dummy_media.capacitor_q(C=C, f_0=F, q_factor=Q)

        Z = 1/(2*np.pi*F*C/Q+1j*C*ntwk.frequency.w)
        ABCD = np.full(ntwk.s.shape, [[1, -1+1j],
                                       [0, 1]])
        ABCD[:,0,1] = Z
        assert_array_almost_equal(ABCD, ntwk.a)

    def test_inductor(self):
        """
        Compare the component values against s-parameters generated by QUCS.
        Test the naming of the network. When circuit is used to connect a
        topology of networks, they should have unique names.
        """
        name = 'inductor,p1nH'
        qucs_ntwk = Network(os.path.join(self.files_dir, name + '.s2p'))
        self.dummy_media.frequency = qucs_ntwk.frequency
        skrf_ntwk = self.dummy_media.inductor(.1e-9, name = name)
        self.assertEqual(qucs_ntwk, skrf_ntwk)
        self.assertEqual(qucs_ntwk.name, skrf_ntwk.name)
        # test vs analytical ABCD parameters of a series inductor
        L = 0.1e-9
        ntwk = self.dummy_media.inductor(L)
        Z = 1j*L*ntwk.frequency.w
        ABCD = np.full(ntwk.s.shape, [[1, -1+1j],
                                       [0, 1]])
        ABCD[:,0,1] = Z
        assert_array_almost_equal(ABCD, ntwk.a)

    def test_inductor_mismatch(self):
        """
        Compare the component values against s-parameters generated by QUCS.
        The mismatch QUCS component is generated by renormalize() method.
        """
        # Config the qucs_ntwk and media
        name = 'inductor,p1nH'
        qucs_ntwk = Network(os.path.join(self.files_dir, name + '.s2p'))
        self.dummy_media.frequency = qucs_ntwk.frequency
        mismatch_z0_tuple = ([25, 50], [25-5j, 50+10j], [25, 50-10j])

        # Test against Y-parameter defination
        def inductor_y_def(media: DefinedGammaZ0, L, **kwargs):
            result = media.match(nports=2, **kwargs)
            w = media.frequency.w
            y = np.zeros(shape=result.s.shape, dtype=complex)
            L = np.array(L)
            y[:, 0, 0] = 1.0 / (1j * w * L)
            y[:, 1, 1] = 1.0 / (1j * w * L)
            y[:, 0, 1] = -1.0 / (1j * w * L)
            y[:, 1, 0] = -1.0 / (1j * w * L)
            result.y = y
            return result

        # Test real and imag port impedance with Y-parameter defination and renormalize()
        for mismatch_z0 in mismatch_z0_tuple:
            qucs_ntwk_copy = qucs_ntwk.copy()
            qucs_ntwk_copy.renormalize(mismatch_z0)
            skrf_ntwk = self.dummy_media.inductor(.1e-9, name = name, z0=mismatch_z0)
            skrf_ntwk_y = inductor_y_def(self.dummy_media, .1e-9, z0=mismatch_z0)
            self.assertEqual(qucs_ntwk_copy, skrf_ntwk)
            assert_array_almost_equal(skrf_ntwk.s, skrf_ntwk_y.s)

        # Config the s_def parameters
        mismatch_z0 = [25, 50-10j]
        s_defs = ("power", "pseudo", "traveling")

        # Test different s_def configurations
        for s_def in s_defs:
            qucs_ntwk_copy = qucs_ntwk.copy()
            qucs_ntwk_copy.renormalize(mismatch_z0, s_def=s_def)
            skrf_ntwk = self.dummy_media.inductor(.1e-9, name=name, z0=mismatch_z0, s_def=s_def)
            skrf_ntwk_y = inductor_y_def(self.dummy_media, .1e-9, z0=mismatch_z0, s_def=s_def)
            self.assertEqual(skrf_ntwk, qucs_ntwk_copy)
            self.assertEqual(skrf_ntwk.s_def, qucs_ntwk_copy.s_def)
            # numerical precision is slightly reduced when using pseudo s_def
            assert_array_almost_equal(skrf_ntwk.s, skrf_ntwk_y.s, decimal={'pseudo': 5}.get(s_def, 6))
            self.assertEqual(skrf_ntwk.s_def, skrf_ntwk_y.s_def)

    def test_shunt_inductor(self):
        """
        Test the naming of the network. When circuit is used to connect a
        topology of networks, they should have unique names.
        """
        name = 'shunt_inductor,p1nH'
        self.dummy_media.frequency = Frequency(1, 1, 1, unit='GHz')
        skrf_ntwk = self.dummy_media.inductor(.1e-9, name = name)
        self.assertEqual(name, skrf_ntwk.name)
        # test vs analytical ABCD parameters of a shunt inductor
        L = 0.1e-9
        ntwk = self.dummy_media.shunt_inductor(L)
        Z = 1j*L*ntwk.frequency.w
        ABCD = np.full(ntwk.s.shape, [[1, 0],
                                       [1-1j, 1]])
        ABCD[:,1,0] = 1/Z
        assert_array_almost_equal(ABCD, ntwk.a)

    def test_inductor_q(self):
        """
        Compare the component values against s-parameters generated by ADS.
        Test the naming of the network. When circuit is used to connect a
        topology of networks, they should have unique names.
        """
        name = 'inductor_q'
        ads_path = os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            'ads'
            )
        ads_ntwk = Network(os.path.join(ads_path, name + '.s2p'))
        self.dummy_media.frequency = ads_ntwk.frequency
        skrf_ntwk = self.dummy_media.inductor_q(L=1.0e-9, f_0=1.0e9, q_factor=30.0, name = name)
        self.assertEqual(ads_ntwk, skrf_ntwk)
        self.assertEqual(ads_ntwk.name, skrf_ntwk.name)
        # test vs analytical ABCD parameters of a series capacitor with q_factor
        L = 0.1e-9
        Q = 30.0
        F = 1.0e9
        ntwk = self.dummy_media.inductor_q(L=L, f_0=F, q_factor=Q)
        rdc = F*0.05*(2 * np.pi)*L/Q
        w_q = 2*np.pi*F
        rq1 = w_q*L/Q
        rq2 = np.sqrt(rq1**2-rdc**2)
        qt = w_q*L/rq2
        rac = ntwk.frequency.w*L/qt

        Z = np.sqrt(rdc**2+rac**2)+1j*ntwk.frequency.w*L
        ABCD = np.full(ntwk.s.shape, [[1, -1+1j],
                                       [0, 1]])
        ABCD[:,0,1] = Z
        assert_array_almost_equal(ABCD, ntwk.a)

    def test_attenuator(self):
        """
        Test the naming of the network. When circuit is used to connect a
        topology of networks, they should have unique names.
        """
        name = 'attenuator,-10dB'
        self.dummy_media.frequency = Frequency(1, 1, 1, unit='GHz')
        skrf_ntwk = self.dummy_media.attenuator(-10, d = 90, unit = 'deg',
                                                name = name)
        self.assertEqual(name, skrf_ntwk.name)

    def test_lossless_mismatch(self):
        """
        Test the naming of the network. When circuit is used to connect a
        topology of networks, they should have unique names.
        """
        name = 'lossless_mismatch,-10dB'
        self.dummy_media.frequency = Frequency(1, 1, 1, unit='GHz')
        skrf_ntwk = self.dummy_media.lossless_mismatch(-10, name = name)
        self.assertEqual(name, skrf_ntwk.name)

    def test_isolator(self):
        """
        Test the naming of the network. When circuit is used to connect a
        topology of networks, they should have unique names.
        """
        name = 'isolator'
        self.dummy_media.frequency = Frequency(1, 1, 1, unit='GHz')
        skrf_ntwk = self.dummy_media.isolator(name = name)
        self.assertEqual(name, skrf_ntwk.name)

    def test_line_floating(self):
        """
        Test the naming of the network. When circuit is used to connect a
        topology of networks, they should have unique names.
        """
        name = 'floating_line'
        self.dummy_media.frequency = Frequency(1, 1, 1, unit='GHz')
        skrf_ntwk = self.dummy_media.line_floating(90, 'deg', name=name)
        self.assertEqual(name, skrf_ntwk.name)

        # Expected to be a 4-port network
        self.assertEqual(skrf_ntwk.nports, 4)

        # zero length case
        s_zero = 1/2 * np.array([[1, 1, 1, -1],
                                 [1, 1, -1, 1],
                                 [1, -1, 1, 1],
                                 [-1, 1, 1, 1]]
                                ).transpose().reshape(-1,4,4)
        for z0 in [50, 1]:
            zero_length = self.dummy_media.line_floating(d=0, z0=z0)
            assert_array_almost_equal(zero_length.s, s_zero)

    def test_line_floating2(self):
        """
        Test against line
        """
        d = 100
        unit = 'mm'
        z0 = 50
        line = self.dummy_media.line(d, unit, z0)
        line_floating = self.dummy_media.line_floating(d, unit, z0)
        gnd = self.dummy_media.short(n_ports=1)
        line_floating = connect(line_floating, 3, gnd, 0)
        line_floating = connect(line_floating, 2, gnd, 0)
        assert_array_almost_equal(line_floating.s, line.s)

    def test_line_floating3(self):
        """
        Test against Y-parameter definition
        """
        d = 100
        unit = 'mm'

        def yparam_ref(media, d, unit, z0, s_def=S_DEF_DEFAULT):
            result = media.match(nports=4, z0=z0, s_def='traveling')

            theta = media.electrical_length(media.to_meters(d=d, unit=unit))

            # From AWR docs on TLINP4.
            y11 = 1 / (z0 * np.tanh(theta))
            y12 = -1 / (z0 * np.sinh(theta))
            y22 = y11
            y21 = y12

            result.y = \
                    np.array([[ y11,  y12, -y11, -y12],
                              [ y21,  y22, -y21, -y22],
                              [-y11, -y12,  y11,  y12],
                              [-y21, -y22,  y21,  y22]]).transpose(2,0,1)
            if media.z0_port is not None:
                result.renormalize(media.z0_port)
            result.renormalize(result.z0, s_def=s_def)
            return result

        for z0 in [50, 1, 100+10j]:
            line = self.dummy_media.line_floating(d, unit, z0)
            ref_line = yparam_ref(self.dummy_media, d, unit, z0)
            assert_array_almost_equal(ref_line.s, line.s)

    def test_scalar_gamma_z0_media(self):
        """
        test ability to create a Media from scalar quantities for gamma/z0
        """
        freq = Frequency(1, 10, 101, unit='GHz')
        a = DefinedGammaZ0(freq, gamma=1j, z0=50)
        self.assertEqual(len(freq), len(a))
        self.assertEqual(len(freq), len(a.gamma))
        self.assertEqual(len(freq), len(a.z0))


    def test_vector_gamma_z0_media(self):
        """
        test ability to create a Media from vector quantities for gamma/z0
        """
        freq = Frequency(1, 10, 101, unit='GHz')
        a = DefinedGammaZ0(freq,
                           gamma = 1j*np.ones(len(freq)) ,
                           z0 =  50*np.ones(len(freq)),
                            )
        self.assertEqual(len(freq), len(a))
        self.assertEqual(len(freq), len(a.gamma))
        self.assertEqual(len(freq), len(a.z0))

    def test_write_csv(self):
        fname = os.path.join(self.files_dir,\
                'out.csv')
        self.dummy_media.write_csv(fname)
        os.remove(fname)


    def test_from_csv(self):
        fname = os.path.join(self.files_dir,\
                'out.csv')
        self.dummy_media.write_csv(fname)
        a_media = DefinedGammaZ0.from_csv(fname)
        self.assertEqual(a_media,self.dummy_media)
        os.remove(fname)


class STwoPortsNetworkTestCase(unittest.TestCase):
    """
    Check that S parameters of media base elements versus theoretical results.
    """
    def setUp(self):
        self.dummy_media = DefinedGammaZ0(
            frequency=Frequency(1, 100, 21, unit='GHz'),
            gamma=1j,
            z0=50,
            )

    def test_s_series_element(self):
        """
        Series elements of impedance Z:

        ○---[Z]---○

        ○---------○

        have S matrix of the form:
        [ Z/Z0 / (Z/Z0 + 2)     2/(Z/Z0 + 2) ]
        [ 2/(Z/Z0 + 2)          Z/Z0 / (Z/Z0 + 2) ]
        """
        R = 1.0  # Ohm
        ntw = self.dummy_media.resistor(R)
        Z0 = self.dummy_media.z0
        S11 = (R/Z0) / (R/Z0 + 2)
        S21 = 2 / (R/Z0 + 2)
        assert_array_almost_equal(ntw.s[:,0,0], S11)
        assert_array_almost_equal(ntw.s[:,0,1], S21)
        assert_array_almost_equal(ntw.s[:,1,0], S21)
        assert_array_almost_equal(ntw.s[:,1,1], S11)

    def test_s_shunt_element(self):
        """
        Shunt elements of admittance Y:

        ○---------○
            |
           [Y]
            |
        ○---------○

        have S matrix of the form:
        [ -Y Z0 / (Y Z0 + 2)     2/(Y Z0 + 2) ]
        [ 2/(Y Z0 + 2)          Z/Z0 / (Y Z0 + 2) ]
        """
        R = 1.0  # Ohm
        ntw = self.dummy_media.shunt(self.dummy_media.resistor(R)**self.dummy_media.short())
        Z0 = self.dummy_media.z0
        S11 = -(1/R*Z0) / (1/R*Z0 + 2)
        S21 = 2 / (1/R*Z0 + 2)
        assert_array_almost_equal(ntw.s[:,0,0], S11)
        assert_array_almost_equal(ntw.s[:,0,1], S21)
        assert_array_almost_equal(ntw.s[:,1,0], S21)
        assert_array_almost_equal(ntw.s[:,1,1], S11)

    def test_s_lossless_line(self):
        """
        Lossless transmission line of characteristic impedance z1, length l
        and wavenumber beta
              _______
        ○-----       -----○
          z0     z1    z0
        ○-----_______-----○

        """
        l = 5.0
        z1 = 30.0
        z0 = self.dummy_media.z0

        ntw = self.dummy_media.line(d=0, unit='m', z0=z0) \
            ** self.dummy_media.line(d=l, unit='m', z0=z1) \
            ** self.dummy_media.line(d=0, unit='m', z0=z0)

        beta = self.dummy_media.beta
        _z1 = z1/z0
        S11 = 1j*(_z1**2 - 1)*np.sin(beta*l) / \
            (2*_z1*np.cos(beta*l) + 1j*(_z1**2 + 1)*np.sin(beta*l))
        S21 = 2*_z1 / \
            (2*_z1*np.cos(beta*l) + 1j*(_z1**2 + 1)*np.sin(beta*l))
        assert_array_almost_equal(ntw.s[:,0,0], S11)
        assert_array_almost_equal(ntw.s[:,0,1], S21)
        assert_array_almost_equal(ntw.s[:,1,0], S21)
        assert_array_almost_equal(ntw.s[:,1,1], S11)

    def test_s_lossy_line(self):
        """
        Lossy transmission line of characteristic impedance Z0, length l
        and propagation constant gamma = alpha + j beta

        ○---------○

        ○---------○

        has ABCD matrix of the form:

        [ cosh(gamma l)       Z0 sinh(gamma l) ]
        [ 1/Z0 sinh(gamma l)  cosh(gamma l) ]
        """


class ABCDTwoPortsNetworkTestCase(unittest.TestCase):
    """
    Check that ABCD parameters of media base elements (such as lumped elements)
    versus theoretical results.
    """
    def setUp(self):
        self.dummy_media = DefinedGammaZ0(
            frequency=Frequency(1, 100, 21, unit='GHz'),
            gamma=1j,
            z0=50 ,
            )

    def test_abcd_series_element(self):
        """
        Series elements of impedance Z:

        ○---[Z]---○

        ○---------○

        have ABCD matrix of the form:
        [ 1  Z ]
        [ 0  1 ]
        """
        R = 1.0  # Ohm
        ntw = self.dummy_media.resistor(R)
        assert_array_almost_equal(ntw.a[:,0,0], 1.0)
        assert_array_almost_equal(ntw.a[:,0,1], R)
        assert_array_almost_equal(ntw.a[:,1,0], 0.0)
        assert_array_almost_equal(ntw.a[:,1,1], 1.0)

    def test_abcd_shunt_element(self):
        """
        Shunt elements of admittance Y:

        ○---------○
            |
           [Y]
            |
        ○---------○

        have ABCD matrix of the form:
        [ 1  0 ]
        [ Y  1 ]
        """
        R = 1.0  # Ohm
        ntw = self.dummy_media.shunt(self.dummy_media.resistor(R)**self.dummy_media.short())
        assert_array_almost_equal(ntw.a[:,0,0], 1.0)
        assert_array_almost_equal(ntw.a[:,0,1], 0.0)
        assert_array_almost_equal(ntw.a[:,1,0], 1.0/R)
        assert_array_almost_equal(ntw.a[:,1,1], 1.0)

    def test_abcd_series_shunt_elements(self):
        """
        Series and Shunt elements of impedance Zs and Zp:

        ○---[Zs]--------○
                 |
                [Zp]
                 |
        ○--------------○
        have ABCD matrix of the form:
        [ 1 + Zs/Zp    Zs ]
        [ 1/Zp          1 ]
        """
        Rs = 2.0
        Rp = 3.0
        serie_resistor = self.dummy_media.resistor(Rs)
        shunt_resistor = self.dummy_media.shunt(self.dummy_media.resistor(Rp) ** self.dummy_media.short())

        ntw = serie_resistor ** shunt_resistor

        assert_array_almost_equal(ntw.a[:,0,0], 1.0+Rs/Rp)
        assert_array_almost_equal(ntw.a[:,0,1], Rs)
        assert_array_almost_equal(ntw.a[:,1,0], 1.0/Rp)
        assert_array_almost_equal(ntw.a[:,1,1], 1.0)

    def test_abcd_thru(self):
        """
        Thru has ABCD matrix of the form:
        [ 1  0 ]
        [ 0  1 ]
        """
        ntw = self.dummy_media.thru()
        assert_array_almost_equal(ntw.a[:,0,0], 1.0)
        assert_array_almost_equal(ntw.a[:,0,1], 0.0)
        assert_array_almost_equal(ntw.a[:,1,0], 0.0)
        assert_array_almost_equal(ntw.a[:,1,1], 1.0)

    def test_abcd_lossless_line(self):
        """
        Lossless transmission line of characteristic impedance Z0, length l
        and wavenumber beta

        ○---------○

        ○---------○

        has ABCD matrix of the form:

        [ cos(beta l)       j Z0 sin(beta l) ]
        [ j/Z0 sin(beta l)  cos(beta l) ]
        """
        l = 5
        z0 = 80
        ntw = self.dummy_media.line(d=l, unit='m', z0=z0)
        beta = self.dummy_media.beta
        assert_array_almost_equal(ntw.a[:,0,0], np.cos(beta*l))
        assert_array_almost_equal(ntw.a[:,0,1], 1j*z0*np.sin(beta*l))
        assert_array_almost_equal(ntw.a[:,1,0], 1j/z0*np.sin(beta*l))
        assert_array_almost_equal(ntw.a[:,1,1], np.cos(beta*l))

    def test_abcd_lossy_line(self):
        """
        Lossy transmission line of characteristic impedance Z0, length l
        and propagation constant gamma = alpha + j beta

        ○---------○

        ○---------○

        has ABCD matrix of the form:

        [ cosh(gamma l)       Z0 sinh(gamma l) ]
        [ 1/Z0 sinh(gamma l)  cosh(gamma l) ]
        """
        l = 5.0
        z0 = 30.0
        alpha = 0.5
        beta = 2.0
        lossy_media = DefinedGammaZ0(
            frequency=Frequency(1, 100, 21, unit='GHz'),
            gamma=alpha + 1j*beta,
            z0=z0
            )
        ntw = lossy_media.line(d=l, unit='m', z0=z0)
        gamma = lossy_media.gamma
        assert_array_almost_equal(ntw.a[:,0,0], np.cosh(gamma*l))
        assert_array_almost_equal(ntw.a[:,0,1], z0*np.sinh(gamma*l))
        assert_array_almost_equal(ntw.a[:,1,0], 1.0/z0*np.sinh(gamma*l))
        assert_array_almost_equal(ntw.a[:,1,1], np.cosh(gamma*l))

class DefinedGammaZ0_s_def(unittest.TestCase):
    """Test various media constructs with complex ports and different s_def"""

    def test_complex_ports(self):
        m = DefinedGammaZ0(
            frequency = Frequency(1, 1, 1, unit='ghz'),
            gamma=1j,
            z0_port = 50,
            z0 = 10+20j,
            )
        self.assertTrue(m.z0.imag != 0)

        # Powerwave short is -Z0.conj()/Z0
        short = m.short(z0=m.z0, s_def='power')
        self.assertTrue(short.s != -1)
        # Should be -1 when converted to traveling s_def
        np.testing.assert_allclose(short.s_traveling, -1)

        short = m.short(z0=m.z0, s_def='traveling')
        np.testing.assert_allclose(short.s, -1)

        short = m.short(z0=m.z0, s_def='pseudo')
        np.testing.assert_allclose(short.s, -1)

        # Mismatches agree with real port impedances
        mismatch_traveling = m.impedance_mismatch(z1=10, z2=50, s_def='traveling')
        mismatch_pseudo = m.impedance_mismatch(z1=10, z2=50, s_def='pseudo')
        mismatch_power = m.impedance_mismatch(z1=10, z2=50, s_def='power')
        np.testing.assert_allclose(mismatch_traveling.s, mismatch_pseudo.s)
        np.testing.assert_allclose(mismatch_traveling.s, mismatch_power.s)

        mismatch_traveling = m.impedance_mismatch(z1=10+10j, z2=50-20j, s_def='traveling')
        mismatch_pseudo = m.impedance_mismatch(z1=10+10j, z2=50-20j, s_def='pseudo')
        mismatch_power = m.impedance_mismatch(z1=10+10j, z2=50-20j, s_def='power')

        # Converting thru to new impedance should give impedance mismatch.
        # The problem is that thru Z-parameters have infinities
        # and renormalization goes through Z-parameters making
        # it very inaccurate.
        thru_traveling = m.thru(s_def='traveling')
        thru_traveling.renormalize(z_new=[10+10j,50-20j])
        thru_pseudo = m.thru(s_def='pseudo')
        thru_pseudo.renormalize(z_new=[10+10j,50-20j])
        thru_power = m.thru(s_def='power')
        thru_power.renormalize(z_new=[10+10j,50-20j])

        np.testing.assert_allclose(thru_traveling.s, mismatch_traveling.s, rtol=1e-3)
        np.testing.assert_allclose(thru_pseudo.s, mismatch_pseudo.s, rtol=1e-3)
        np.testing.assert_allclose(thru_power.s, mismatch_power.s, rtol=1e-3)

if __name__ == '__main__':
    unittest.main()



import os
import unittest

import numpy as np
from numpy.testing import assert_allclose, assert_almost_equal, assert_array_equal

import skrf as rf


class QfactorTests(unittest.TestCase):
    """ Q-factor class tests.

    References
    ----------
    "Q-factor Measurement by using a Vector Network Analyser",
        A. P. Gregory, National Physical Laboratory Report MAT 58 (2021)
        https://eprintspublications.npl.co.uk/9304/
    """

    def setUp(self):
        """
        Q-factor tests initalizer.
        """
        self.test_dir = os.path.dirname(os.path.abspath(__file__)) + "/"
        self.ntwk_2port = rf.data.ring_slot
        self.ntwk_1port = rf.data.ring_slot_meas

    def csv_file_example_to_network(self, file: str) -> rf.Network:
        """Convert the S-parameter txt file to Network.

        Parameters
        ----------
        file : str
            file path.

        Returns
        -------
        ntwk : rf.Network (1-port)

        """
        # Load frequency and S-parameter data from a file.
        try:
            f, s_re, s_im = np.loadtxt(file, comments="%", unpack=True)
        except ValueError:
            f, s_re, s_im, s_abs, s_mag = np.loadtxt(file, comments="%", unpack=True)

        s = s_re + 1j * s_im
        freq = rf.Frequency.from_f(f, unit='GHz')
        return rf.Network(s=s, frequency=freq)

    def test_constructor(self):
        """
        Test the Qfactor() constructor.
        """
        # constructor tests
        _Q1 = rf.Qfactor(self.ntwk_1port, res_type='reflection')
        _Q2 = rf.Qfactor(self.ntwk_1port, res_type='reflection', Q_L0=3)
        _Q3 = rf.Qfactor(self.ntwk_1port, res_type='reflection', f_L0=85e9)
        _Q3 = rf.Qfactor(self.ntwk_1port, res_type='reflection', Q_L0=3, f_L0=85e9)

    def test_exceptions(self):
        "Test the raised exceptions."

        # Passing a 2-port Network raises a ValueError
        self.assertRaises(ValueError, rf.Qfactor, self.ntwk_2port, 'reflection')

        # Uncorrect resonance type raises a ValueError
        self.assertRaises(ValueError, rf.Qfactor, self.ntwk_1port, 'dummy')

        # Asking for fitted S-param and Network without prior fit raises a ValueError
        _Q = rf.Qfactor(self.ntwk_1port, res_type='reflection')
        self.assertRaises(ValueError, _Q.fitted_s)
        self.assertRaises(ValueError, _Q.fitted_network)

    def test_NLQFIT6(self):
        """
        Fit FL and QL to transmission (S21) data by using the NLQFIT6 algorithm.

        References
        ----------
        Test data is read from file Figure6b.txt (as used in Figure 6(b) of MAT 58).

        """
        # File 'Figure6b.txt' contains S21 data for Fig. 6(b) in MAT 58
        ntwk = self.csv_file_example_to_network(self.test_dir + "qfactor_data/Figure6b.txt")

        Q = rf.Qfactor(ntwk, res_type='transmission', verbose=True)

        # Test against expected solutions
        assert_almost_equal(Q._a, 0.8104 - 1.6928j, decimal=4)
        assert_almost_equal(Q._b, 0.0077 - 0.0071j, decimal=4)
        assert_almost_equal(Q.Q_L, 7440.848, decimal=3)

        # Optimised weighted fit --> result vector
        res = Q.fit(method="NLQFIT6")

        # Test against expected solutions
        assert_allclose(res.f_L, 3.987848e9)
        assert_almost_equal(res.Q_L, 7454, decimal=0)
        assert_almost_equal(res.RMS_Error, 0.00001216)
        assert_array_equal(res.f_L, Q.f_L)
        assert_array_equal(res.Q_L, Q.Q_L)

        # Now calculate unloaded Q-factor and some other useful quantities.
        # Reciprocal of |S21| of a thru in place of resonator
        scaling_factor_A = 1 / 0.874  # 1/|S21_thru|
        Q0 = Q.Q_unloaded(res, scaling_factor_A)
        cal_diam, cal_gamma_V, cal_gamma_T = Q.Q_circle(res, scaling_factor_A)

        # Q-factor of uncoupled two-port resonator (unloaded Q-factor)
        assert_almost_equal(Q0, 7546, decimal=0)
        assert_almost_equal(cal_diam, 0.0121, decimal=4)
        # S21 detuned = leakage vector
        assert_almost_equal(cal_gamma_V, -0.00008895 + 0.00003852j)
        # S21 tuned
        assert_almost_equal(cal_gamma_T, 0.00849357 - 0.00845349j)

    def test_NLQFIT6_2(self):
        """
        Fit FL and QL to transmission (S21) data by using the NLQFIT6 algorithm.

        From a two-port superconducting absorption resonator.

        References
        ----------
        Test data is read from file Figure27.txt (as used in Figure 27 of MAT 58).

        """
        ntwk = self.csv_file_example_to_network(self.test_dir + "qfactor_data/Figure27.txt")

        Q = rf.Qfactor(ntwk, res_type='absorption', verbose=True)

        # Test against expected solutions
        assert_almost_equal(Q._a, -17072.3098 + 9047.0761j, decimal=4)
        assert_almost_equal(Q._b, 0.0063 + 0.0168j, decimal=4)

        # Q.tol = 1.0e-5 * np.argmax(np.abs(Q.s))

        # Step 2: Optimised weighted fit --> result vector
        res = Q.fit(method="NLQFIT6")

        assert_allclose(res.f_L, 6.07225567e9)
        assert_almost_equal(res.Q_L, 56019.85, decimal=0)
        assert_almost_equal(res.weighting_ratio, 4.714, decimal=3)
        assert_almost_equal(res.RMS_Error, 0.01394722)
        assert_allclose(Q.Q_L, res.Q_L)
        assert_allclose(Q.f_L, res.f_L)

        Q0 = Q.Q_unloaded(res)
        cal_diam, cal_gamma_V, cal_gamma_T = Q.Q_circle(res)

        assert_allclose(Q0, 1846782, rtol=1/100)
        assert_allclose(cal_diam, 0.970, rtol=1/100)
        assert_almost_equal(2.0 * res.f_L / res.Q_L/1e9, 0.00021678942580071302, decimal=10)



    def test_NLQFIT7(self):
        """
        Fit FL and QL to reflection (S11) data by using the NLQFIT7 algorithm.

        References
        ----------
        Test data is read from file Table6c27.txt (as used in Figure 16 and
        Table 6(c) of MAT 58).

        """
        ntwk = self.csv_file_example_to_network(self.test_dir + "qfactor_data/Table6c27.txt")

        Q = rf.Qfactor(ntwk, res_type='reflection', verbose=True)
        # Expected results after initial fit
        assert_almost_equal(Q._a, 760.9731 + 67.7804j, decimal=4)
        assert_almost_equal(Q._b, 0.0609 - 0.6432j, decimal=4)
        assert_almost_equal(Q.Q_L, 779.068, decimal=3)

        # Expected results after fit
        res = Q.fit(method='NLQFIT7')
        # Fitted length of uncalibrated line [m]
        assert_almost_equal(-res.m7a*rf.c/(4.0*np.pi*1.3), 57.47056249053462e-3)
        assert_allclose(Q.f_L, 3.65293800e9)
        assert_almost_equal(Q.Q_L, 708, decimal=0)

        # Unloaded Q-factor and some other useful quantities.
        print("Q-factor of unloaded one-port resonator by Method 1:")
        print("Assumes attenuating uncalibrated line")
        Q0 = Q.Q_unloaded(res)
        cal_diam, cal_gamma_V, cal_gamma_T = Q.Q_circle(res)

        # Test against expected solutions
        assert_almost_equal(Q0, 862, decimal=0)
        assert_almost_equal(cal_diam, 0.3573, decimal=4)
        assert_almost_equal(cal_gamma_V, 0.09084890 - 0.99586469j)  # S11 detuned
        assert_almost_equal(cal_gamma_T, 0.05878773 - 0.64003179j)  # S11 tuned

        print("Q-factor of unloaded one-port resonator by Method 2:")
        print("Scaling factor A = 1.0 (assume no attenuation in uncalibrated line)")
        Q2 = rf.Qfactor(ntwk, res_type='reflection_method2', verbose=True)
        res2 = Q2.fit(method='NLQFIT7')

        Q0_2 = Q.Q_unloaded(res2, A=1)
        cal_diam2, cal_gamma_V2, cal_gamma_T2 = Q.Q_circle(res2, A=1)

        # Test against expected solutions
        assert_almost_equal(Q0_2, 862, decimal=0)
        assert_almost_equal(res2.weighting_ratio, 28.317, decimal=3)
        assert_almost_equal(res2.RMS_Error, 0.00145957)
        assert_almost_equal(cal_diam2, 0.3573, decimal=2)
        assert_almost_equal(cal_gamma_V2, 0.08996562 - 0.98618239j)
        assert_almost_equal(cal_gamma_T2, 0.05821616 - 0.63380908j)

    def test_NLQFIT8(self):
        """
        Fits to transmission (S21) data by using the NLQFIT8 algorithm.

        frequency-dependent leakage

        References
        ----------
        Test data is read from file Figure23.txt (shown in Figure 23 of MAT 58)

        """
        ntwk = self.csv_file_example_to_network(self.test_dir + "qfactor_data/Figure23.txt")

        Q = rf.Qfactor(ntwk, res_type='transmission')

        # # De-embed cables
        # N = len(Q.f)
        ncablelen = 1.2  # root_eps * cable length in metres
        # D =  ntwk.s[:,0,0]* np.exp(-1j * 2*np.pi *Q.f * ncablelen / rf.c)

        # ntwk2 = ntwk.copy()
        # ntwk2.s[:,0,0] = D

        # piece of transmission line to deembbed
        gamma = 0 - 1j*ntwk.frequency.w * ncablelen / rf.c
        coax = rf.media.DefinedGammaZ0(frequency=ntwk.frequency, gamma=gamma)
        line = coax.line(1/2, unit='m')
        ntwk2 = line.inv ** ntwk

        # Find peak in |S21| - this is used to give initial value of freq.
        # Tol is 1.0E-5 * |S21| at peak.
        Mg = np.abs(ntwk2.s).squeeze()
        index_max = np.argmax(Mg)
        Fseed = Q.f[index_max]

        # Set Qseed: An order-of-magnitude estimate for Q-factor
        mult = 5.0  # Not critical. A value of around 5.0 will work well for initial and optimised fits (Section 2.6).
        Qseed = mult * Fseed / (Q.f[-1] - Q.f[0])

        Q = rf.Qfactor(ntwk2, res_type='transmission', Q_L0=Qseed, f_L0=Fseed)
        assert_almost_equal(Q._a, 8.9408 + 2.1298j, decimal=4)
        assert_almost_equal(Q._b, 0.0054 - 0.0045j, decimal=4)
        assert_almost_equal(Q.Q_L, 4664.2418, decimal=3)

        # Step 2: Optimised weighted fit
        res = Q.fit(method="NLQFIT8", loop_plan="fwfwfwc")

        # Unloaded Q-factor and some other useful quantities.
        # Reciprocal of |S21| of a thru in place of resonator
        scaling_factor_A = 1 / 0.949  # 1/|S21_thru|
        Q0 = Q.Q_unloaded(res, scaling_factor_A)
        cal_diam, cal_gamma_V, cal_gamma_T = Q.Q_circle(res, scaling_factor_A)

        # Test against expected solutions
        assert_allclose(Q.f_L, 9.76015571e9)
        assert_almost_equal(Q.Q_L, 4760.04, decimal=0)
        assert_almost_equal(Q0, 4789.49, decimal=0)
        assert_almost_equal(res.weighting_ratio, 5.078, decimal=3)
        assert_almost_equal(res.RMS_Error, 0.00000910)
        assert_almost_equal(cal_diam, 0.0061, decimal=2)

    def test_Q_unloaded(self):
        """Test unloaded Q factor method."""
        Q = rf.Qfactor(self.ntwk_1port, res_type='reflection')
        res = Q.fit()
        self.assertRaises(ValueError, Q.Q_unloaded, A='dummy')
        self.assertRaises(ValueError, Q.Q_unloaded, A=1j)
        self.assertRaises(ValueError, Q.Q_unloaded, res, A='dummy')
        self.assertRaises(ValueError, Q.Q_unloaded, res, A=1j)

        # passing of not the fitted results after fit should be the same
        self.assertEqual(Q.Q_unloaded(res), Q.Q_unloaded())
        # passing a different solution should lead to different values
        res2 = Q.fit(method="NLQFIT7")
        self.assertNotEqual(Q.Q_unloaded(res), Q.Q_unloaded(res2))

    def test_Q_circle(self):
        """Test Q-circle method."""
        Q = rf.Qfactor(self.ntwk_1port, res_type='reflection')
        res = Q.fit(method="NLQFIT6")
        self.assertRaises(ValueError, Q.Q_circle, A='dummy')
        self.assertRaises(ValueError, Q.Q_circle, A=1j)
        self.assertRaises(ValueError, Q.Q_circle, res, A='dummy')
        self.assertRaises(ValueError, Q.Q_circle, res, A=1j)

        # passing of not the fitted results after fit should be the same
        self.assertEqual(Q.Q_circle(res), Q.Q_circle())
        # passing a different solution should lead to different values
        res2 = Q.fit(method="NLQFIT7")
        self.assertNotEqual(Q.Q_circle(res), Q.Q_circle(res2))

    def test_f_L(self):
        "Test resonant frequency values."
        # expected values
        f_L_expected = self.ntwk_2port.f[np.argmin(self.ntwk_2port.s11.s_mag)]
        f_L_expected_scaled = f_L_expected/self.ntwk_2port.frequency.multiplier
        # fitted values
        Q = rf.Qfactor(self.ntwk_2port.s11, res_type='reflection')
        # before the fit, warnings should be raised
        with self.assertWarns(Warning):
            # the resonance frequency corresponds to min value before fitting
            assert_almost_equal(Q.f_L, f_L_expected)
            assert_almost_equal(Q.f_L_scaled, f_L_expected_scaled)
        # NB: after the fit this should not be the case anymore (slight deviation)

    def test_BW(self):
        "Test bandwidth values."
        Q = rf.Qfactor(self.ntwk_2port.s11, res_type='reflection')
        # before the fit, warnings should be raised
        with self.assertWarns(Warning):
            Q.BW
            Q.BW_scaled



"""
distributedCircuit (:mod:`skrf.media.distributedCircuit`)
============================================================

A transmission line mode defined in terms of distributed impedance and admittance values.

.. autosummary::
   :toctree: generated/

   DistributedCircuit

"""
from __future__ import annotations

from typing import TYPE_CHECKING

from numpy import imag, real, sqrt

from ..constants import NumberLike
from .media import DefinedGammaZ0, Media

if TYPE_CHECKING:
    from ..frequency import Frequency


class DistributedCircuit(Media):
    r"""
    A transmission line mode defined in terms of distributed impedance and admittance values.

    Parameters
    ------------
    frequency : :class:`~skrf.frequency.Frequency` object
        frequency band of the media
    z0_port : number, array-like, or None
        `z0_port` is the port impedance for networks generated by the media.
        If `z0_port` is not None, the networks generated by the media are
        renormalized (or in other words embedded) from the characteristic
        impedance z0 of the media to `z0_port`.
        Else if `z0_port` is None, the networks port impedances will be the raw
        characteristic impedance z0 of the media.
        (Default is None)
    z0_override : number, array-like, or None
        `z0_override` override the characteristic impedance for the media.
        If `z0_override` is not None, the networks generated by the media have
        their characteristic impedance `z0` overrided by `z0_override`.
        (Default is None)
    z0 : number, array-like, or None
        deprecated parameter, alias to `z0_override` if `z0_override` is None.
        Emmit a deprecation warning.
    C : number, or array-like
            distributed capacitance, in F/m
    L : number, or array-like
            distributed inductance, in H/m
    R : number, or array-like
            distributed resistance, in Ohm/m
    G : number, or array-like
            distributed conductance, in S/m


    Notes
    -----
    if C,I,R,G are vectors they should be the same length

    :class:`DistributedCircuit` is `Media` object representing a
    transmission line mode defined in terms of distributed impedance
    and admittance values.

    A `DistributedCircuit` may be defined in terms
    of the following attributes:

    ================================  ================  ================
    Quantity                          Symbol            Property
    ================================  ================  ================
    Distributed Capacitance           :math:`C^{'}`     :attr:`C`
    Distributed Inductance            :math:`L^{'}`     :attr:`L`
    Distributed Resistance            :math:`R^{'}`     :attr:`R`
    Distributed Conductance           :math:`G^{'}`     :attr:`G`
    ================================  ================  ================


    The following quantities may be calculated, which are functions of
    angular frequency (:math:`\omega`):

    ===================================  ==================================================  ========================
    Quantity                             Symbol                                              Property
    ===================================  ==================================================  ========================
    Distributed Impedance                :math:`Z^{'} = R^{'} + j \omega L^{'}`              :attr:`Z`
    Distributed Admittance               :math:`Y^{'} = G^{'} + j \omega C^{'}`              :attr:`Y`
    ===================================  ==================================================  ========================


    The properties which define their wave behavior:

    ===================================  ============================================  ==============================
    Quantity                             Symbol                                        Method
    ===================================  ============================================  ==============================
    Characteristic Impedance             :math:`Z_0 = \sqrt{ \frac{Z^{'}}{Y^{'}}}`     :func:`Z0`
    Propagation Constant                 :math:`\gamma = \sqrt{ Z^{'}  Y^{'}}`         :func:`gamma`
    ===================================  ============================================  ==============================

    Given the following definitions, the components of propagation
    constant are interpreted as follows:

    .. math::

        +\Re e\{\gamma\} = \text{attenuation}

        -\Im m\{\gamma\} = \text{forward propagation}


    See Also
    --------
    from_media

    """

    def __init__(self, frequency: Frequency | None = None,
                 z0_port: NumberLike | None = None,
                 z0_override: NumberLike | None = None,
                 z0: NumberLike | None = None,
                 C: NumberLike = 90e-12, L: NumberLike = 280e-9,
                 R: NumberLike = 0, G: NumberLike = 0,
                *args, **kwargs):
        Media.__init__(self, frequency = frequency,
                       z0_port = z0_port, z0_override = z0_override, z0 = z0)

        self.C, self.L, self.R, self.G = C,L,R,G


    def __str__(self) -> str:
        f=self.frequency
        try:
            output = (
                f'Distributed Circuit Media.  {f.f_scaled[0]}-{f.f_scaled[-1]} {f.unit}.  {f.npoints} points'
                f'\nL\'= {self.L:.2f}, C\'= {self.C:.2f},R\'= {self.R:.2f}, G\'= {self.G:.2f}, ')
        except(TypeError):
            output = (
                f'Distributed Circuit Media.  {f.f_scaled[0]}-{f.f_scaled[-1]} {f.unit}.  {f.npoints} points'
                f'\nL\'= {self.L:.2f}.., C\'= {self.C:.2f}..,R\'= {self.R:.2f}.., G\'= {self.G:.2f}.., ')
        return output

    def __repr__(self) -> str:
        return self.__str__()

    @classmethod
    def from_media(cls, my_media: Media, *args, **kwargs) -> Media:
        """
        Initializes a `DistributedCircuit` from an existing
        :class:`~skrf.media.media.Media` instance.

        Parameters
        ----------
        my_media : :class:`~skrf.media.media.Media` instance.
            the media object

        See Also
        --------
        :class:`~skrf.media.media.Media`
        """

        w  =  my_media.frequency.w
        gamma = my_media.gamma
        z0 = my_media.z0
        z0_port = my_media.z0_port

        Y = gamma/z0
        Z = gamma*z0
        G,C = real(Y), imag(Y)/w
        R,L = real(Z), imag(Z)/w
        return cls(frequency = my_media.frequency,
                   z0_port = z0_port, C=C, L=L, R=R, G=G, **kwargs)

    @classmethod
    def from_csv(self, *args, **kw):
        """
        Create a `DistributedCircuit` from numerical values stored in a csv file.

        The csv file format must be written by the function :func:`write_csv`,
        or similar method which produces the following format::

            f[$unit], Re(z0), Im(z0), Re(gamma), Im(gamma), Re(z0_port), Im(z0_port)
            1, 1, 1, 1, 1, 1, 1
            2, 1, 1, 1, 1, 1, 1
            .....

        See Also
        --------
        write_csv
        """
        d = DefinedGammaZ0.from_csv(*args,**kw)
        return self.from_media(d)

    @property
    def Z(self) -> NumberLike:
        r"""
        Distributed Impedance, :math:`Z^{'}`.

        Defined as

        .. math::

                Z^{'} = R^{'} + j \omega L^{'}

        Returns
        -------
        Z : np.ndarray
            Distributed impedance in units of ohm/m
        """
        w  = self.frequency.w
        Z = self.R + 1j*w*self.L
        # Avoid divide by zero.
        # Needs to be imaginary to avoid all divide by zeros in the media class.
        Z[Z.imag == 0] += 1j*1e-12
        return Z

    @property
    def Y(self) -> NumberLike:
        r"""
        Distributed Admittance, :math:`Y^{'}`.

        Defined as

        .. math::

                Y^{'} = G^{'} + j \omega C^{'}

        Returns
        -------
        Y : np.ndarray
            Distributed Admittance in units of S/m
        """

        w  = self.frequency.w
        Y = self.G + 1j*w*self.C
        # Avoid divide by zero.
        # Needs to be imaginary to avoid all divide by zeros in the media class.
        Y[Y.imag == 0] += 1j*1e-12
        return Y

    @property
    def z0_characteristic(self) -> NumberLike:
        r"""
        Characteristic Impedance, :math:`z_0`

        .. math::

                Z_0 = \sqrt{ \frac{Z^{'}}{Y^{'}}}

        Returns
        -------
        z0_characteristic : np.ndarray
            Characteristic Impedance in units of ohms
        """
        return sqrt(self.Z/self.Y)

    @property
    def gamma(self) -> NumberLike:
        r"""
        Propagation Constant, :math:`\gamma`.

        Defined as,

        .. math::

                \gamma =  \sqrt{ Z^{'}  Y^{'}}

        Returns
        -------
        gamma : np.ndarray
                Propagation Constant,

        Note
        ----
        The components of propagation constant are interpreted as follows:

        * positive real(gamma) = attenuation
        * positive imag(gamma) = forward propagation
        """
        return sqrt(self.Z*self.Y)



import re
from enum import Enum

import pytest

from skrf.vi import validators
from skrf.vi.validators import ValidationError


@pytest.mark.parametrize(
    "bounds,test_in,expected",
    [
        ((None, None), 0, 0),
        ((None, None), 0.0, 0),
        ((None, None), "0", 0),
        ((None, None), "-1", -1),
        ((None, None), "1", 1),
        ((None, None), "100000", 100000),
        ((-1, None), "-1", -1),
        ((None, 1), "0", 0),
        ((-1, 1), 1, 1),
        ((0, 100), "1", 1),
    ],
)
def test_int_validator(bounds, test_in, expected):
    v = validators.IntValidator(*bounds)
    assert v.validate_input(test_in) == expected


@pytest.mark.parametrize(
    "bounds,test_in",
    [
        ((-1, 1), 2),
        ((-1, 1), -2.0),
        ((-1, 1), "-2"),
        ((-1, None), -2),
        ((None, 1), 2),
        ((None, 1), 2),
    ],
)
def test_int_validator_out_of_bounds(bounds, test_in):
    v = validators.IntValidator(*bounds)
    with pytest.raises(ValidationError):
        v.validate_input(test_in)


@pytest.mark.parametrize(
    "bounds,test_in,expected",
    [
        ((None, None), 0.0, 0.0),
        ((None, None), 0, 0.0),
        ((None, None), "0", 0.0),
        ((None, None), "0.", 0.0),
        ((None, None), "-1", -1.0),
        ((None, None), "1.0", 1.0),
        ((None, None), "2.5", 2.5),
        ((None, None), "-2.5", -2.5),
        ((-1, None), "0", 0),
        ((None, 1), "0", 0),
        ((-0.5, 0.5), "0.25", 0.25),
    ],
)
def test_float_validator(bounds, test_in, expected):
    v = validators.FloatValidator(*bounds)
    assert v.validate_input(test_in) == expected


@pytest.mark.parametrize(
    "bounds,test_in",
    [
        ((-1, 1), 2),
        ((-1, 1), -2),
        ((-1, 1), "-2.0"),
        ((-1, None), -2),
        ((None, 1), 2),
    ],
)
def test_float_validator_out_of_bounds(bounds, test_in):
    v = validators.FloatValidator(*bounds)
    with pytest.raises(ValidationError):
        v.validate_input(test_in)


@pytest.mark.parametrize(
    "test_in,expected",
    [
        ("100", 100),
        ("1hz", 1),
        ("1hZ", 1),
        ("1Hz", 1),
        ("1HZ", 1),
        ("1 Hz", 1),
        ("1 kHz", 1000),
        ("1 MHz", 1_000_000),
        ("1 GHz", 1_000_000_000),
        ("1.5 GHz", 1_500_000_000),
        ("1 kHz", 1000),
    ],
)
def test_freq_validator(test_in, expected):
    v = validators.FreqValidator()
    assert v.validate_input(test_in) == expected


def test_enum_validator():
    class Foo(Enum):
        A = "A"
        B = "B"

    v = validators.EnumValidator(Foo)
    assert v.validate_input("A") == "A"
    assert v.validate_input("B") == "B"
    assert v.validate_input(Foo.A) == "A"
    assert v.validate_input(Foo.B) == "B"

    with pytest.raises(ValidationError):
        v.validate_input("C")

    assert v.validate_output('A') == Foo.A
    assert v.validate_output('B') == Foo.B
    with pytest.raises(ValidationError):
        v.validate_output("C")


def test_set_validator():
    v = validators.SetValidator([1, 2])
    assert v.validate_input(1) == 1
    assert v.validate_input(2) == 2
    with pytest.raises(ValidationError):
        v.validate_input(3)

    with pytest.raises(ValueError):
        validators.SetValidator([1, '2'])

def test_dict_validator():
    arg_str = "{a},{b}"
    resp_pat = r"(?P<a>\d),(?P<b>\d)"
    v = validators.DictValidator(arg_str, resp_pat)

    assert v.validate_input({"a":1, "b":2}) == "1,2"
    assert v.validate_output('1,2') == {"a":'1', "b":'2'}

    with pytest.raises(ValidationError):
        v.validate_input({"a":1})

    with pytest.raises(ValidationError):
        v.validate_output('1,2,3')

    v = validators.DictValidator(arg_str, re.compile(r"(?P<a>\d),(?P<b>\d)"))
    assert v.validate_input({"a":1, "b":2}) == "1,2"



"""
coaxial (:mod:`skrf.media.coaxial`)
============================================================

A coaxial transmission line defined from its electrical or geometrical/physical properties

.. autosummary::
   :toctree: generated/

   Coaxial

"""

from __future__ import annotations

from typing import TYPE_CHECKING

from numpy import array, exp, expm1, imag, log, size, sqrt
from scipy.constants import c, epsilon_0, mu_0, pi

from ..constants import INF, NumberLike
from ..mathFunctions import db_2_np, db_per_100feet_2_db_per_100meter, feet_2_meter
from ..tlineFunctions import skin_depth, surface_resistivity
from .distributedCircuit import DistributedCircuit
from .media import DefinedGammaZ0, Media

if TYPE_CHECKING:
    from ..frequency import Frequency


class Coaxial(DistributedCircuit, Media):
    """
    A coaxial transmission line defined in terms of its inner/outer
    diameters and permittivity.

    Parameters
    ----------
    frequency : :class:`~skrf.frequency.Frequency` object
        Frequency band.
    z0_port : number, array-like, or None
        `z0_port` is the port impedance for networks generated by the media.
        If `z0_port` is not None, the networks generated by the media are
        renormalized (or in other words embedded) from the characteristic
        impedance z0 of the media to `z0_port`.
        Else if `z0_port` is None, the networks port impedances will be the raw
        characteristic impedance z0 of the media.
        (Default is None)
    z0_override : number, array-like, or None
        `z0_override` override the characteristic impedance for the media.
        If `z0_override` is not None, the networks generated by the media have
        their characteristic impedance `z0` overrided by `z0_override`.
        (Default is None)
    z0 : number, array-like, or None
        deprecated parameter, alias to `z0_override` if `z0_override` is None.
        Emmit a deprecation warning.
    Dint : number, or array-like
        inner conductor diameter, in m
    Dout : number, or array-like
        outer conductor diameter, in m
    epsilon_r : number, or array-like, optional
        relative permittivity of the dielectric medium.
        Default is 1.
    tan_delta : number, or array-like, optional
        loss tangent of the dielectric medium.
        Default is 0.
    sigma : number, or array-like, optional.
        conductors electrical conductivity, in S/m.
        Default is infinity (non lossy metal).


    TODO : different conductivity in case of different conductor kind

    Note
    ----
    Dint, Dout, epsilon_r, tan_delta, sigma can all be vectors as long
    as they are the same length

    References
    ----------
    .. [#] Pozar, D.M.; , "Microwave Engineering", Wiley India Pvt. Limited, 1 sept. 2009

    """
    ## CONSTRUCTOR
    def __init__(self, frequency: Frequency | None = None,
                 z0_port: NumberLike | None = None,
                 z0_override: NumberLike | None = None,
                 z0: NumberLike | None = None,
                 Dint: NumberLike = .81e-3, Dout: NumberLike = 5e-3,
                 epsilon_r: NumberLike = 1, tan_delta: NumberLike = 0,
                 sigma: NumberLike = INF,
                 *args, **kwargs):
        Media.__init__(self, frequency = frequency,
                       z0_port = z0_port, z0_override = z0_override, z0 = z0)

        self.Dint, self.Dout = Dint,Dout
        self.epsilon_r, self.tan_delta, self.sigma = epsilon_r, tan_delta, sigma
        self.epsilon_prime = epsilon_0*self.epsilon_r
        self.epsilon_second = epsilon_0*self.epsilon_r*self.tan_delta

    @classmethod
    def from_attenuation_VF(cls, frequency: Frequency | None = None,
                            z0_port: NumberLike | None = None, z0: float = 50,
                         att=0, unit='dB/m', VF=1) -> Media:
        """
        Init from electrical properties of the line: attenuation and velocity factor.

        Attenuation can be expressed in dB/m, dB/100m, dB/ft, dB/100ft, Neper/m or Neper/ft.
        Default unit is dB/m. A different unit is set by the `unit` parameter.

        Parameters
        ----------
        frequency : :class:`~skrf.frequency.Frequency` object

        z0_port : number, array-like, or None
            `z0_port` is the port impedance for networks generated by the media.
            If `z0_port` is not None, the networks generated by the media are
            renormalized (or in other words embedded) from the characteristic
            impedance z0 of the media to `z0_port`.
            Else if `z0_port` is None, the networks port impedances will be the raw
            characteristic impedance z0 of the media.
            (Default is None)
        z0 : number
            desired characteristic impedance
        att : number, or array-like. optional
            Attenuation of the coaxial line. The default is 0.
            If passed as an array, should be of same size than the frequency.
        unit : string, optional
            Unit of the attenuation. Can be: 'dB/m', dB/100m'', 'dB/ft', 'dB/100ft',
            'Neper/m' or 'Neper/ft' (or 'Np/m', 'Np/ft'). The default is 'dB/m'.
        VF : number, or array-like. optional
            Velocity Factor VF [VF]_. The default is 1.
            If passed as an array, should be of same size than the frequency.

        Returns
        -------
        media : :class:`~skrf.media.media.Media`

        References
        ----------
        .. [VF] : https://www.microwaves101.com/encyclopedias/light-phase-and-group-velocities

        """
        # test size of parameters
        if size(array(att, dtype="object")) not in (1, size(frequency.f)):
            raise ValueError('Attenuation should be scalar or of same size that the frequency.')

        # create gamma
        if unit in ('dB/m', 'db/m'):
            alpha = db_2_np(att)
        elif unit in ('dB/100m', 'db/100m'):
            alpha = db_2_np(att/100)
        elif unit in ('dB/ft', 'dB/feet'):
            alpha = db_2_np(att/feet_2_meter())
        elif unit in ('dB/100ft', 'dB/100feet'):
            alpha = db_2_np(db_per_100feet_2_db_per_100meter(att)/100)
        elif unit in ('Np/m', 'np/m', 'n/m'):
            alpha = att
        elif unit in ('Neper/feet', 'Neper/ft',
                      'Np/feet', 'Np/ft',
                      'np/feet', 'np/ft',
                      'N/feet', 'N/ft'):
            alpha = att/feet_2_meter()
        else:
            raise ValueError('Incorrect attenuation unit. Please see documentation. ', unit)

        beta = 2 * pi * frequency.f / c / VF

        gamma = alpha + 1j*beta

        # return media object from z0 and gamma
        return DefinedGammaZ0(frequency=frequency, gamma=gamma,
                                    z0_port=z0_port, z0=z0)

    @classmethod
    def from_Z0_Dout(cls, frequency: Frequency | None = None,
                     z0_port: NumberLike | None = None, z0: float = 50,
                     epsilon_r: NumberLike = 1, Dout: NumberLike = 5e-3,
                     **kw) -> Media:
        """
        Init from characteristic impedance and outer diameter.

        Parameters
        ----------
        frequency : :class:`~skrf.frequency.Frequency` object

        z0_port : number, array-like, or None
            `z0_port` is the port impedance for networks generated by the media.
            If `z0_port` is not None, the networks generated by the media are
            renormalized (or in other words embedded) from the characteristic
            impedance z0 of the media to `z0_port`.
            Else if `z0_port` is None, the networks port impedances will be the raw
            characteristic impedance z0 of the media.
            (Default is None)
        z0 : number
            desired characteristic impedance
        Dout : number, or array-like
            outer conductor diameter, in m
        epsilon_r=1 : number, or array-like
            relative permittivity of the dielectric medium
        **kw :
            passed to __init__

        Returns
        -------
        media : :class:`~skrf.media.media.Media`
        """
        ep= epsilon_0*epsilon_r

        if imag(z0) !=0:
            raise NotImplementedError()

        b = Dout/2.
        b_over_a = exp(2*pi*z0*sqrt(ep/mu_0))
        a = b/b_over_a
        Dint = 2*a
        return cls(frequency=frequency, z0_port = z0_port, Dint=Dint, Dout=Dout,
                    epsilon_r=epsilon_r, **kw)


    @property
    def Rs(self) -> NumberLike:
        """
        Surface resistivity in Ohm/area.

        Returns
        -------
        Rs : number or array
            surface resistivity

        """
        f  = self.frequency.f
        rho = 1./self.sigma
        mu_r =1
        return surface_resistivity(f=f,rho=rho, mu_r=mu_r)

    @property
    def a(self) -> NumberLike:
        """
        Inner radius of the coaxial line

        Returns
        -------
        a : float
            Inner radius

        """
        return self.Dint/2.

    @property
    def b(self) -> NumberLike:
        """
        Outer radius of the coaxial line

        Returns
        -------
        b : float
            Outer radius
        """
        return self.Dout/2.

    # derivation of distributed circuit parameters
    @property
    def R(self) -> NumberLike:
        """
        Distributed resistance R, in Ohm/m.
        See [#]_ for more information.

        Returns
        -------
        R : number, or array-like
            distributed resistance, in Ohm/m

        References
        -------

        .. [#] https://www.microwaves101.com/encyclopedias/a-more-exact-coax-attenuation-solution

        """
        rho = 1/self.sigma
        delta = skin_depth(f=self.frequency.f, rho=rho, mu_r=1)
        # Avoid infinites at DC
        delta[delta > 1e6] = 1e6

        Rin = rho / (2*pi*delta*self.a + 2*pi*delta**2*expm1(-self.a/delta))
        # Outer shield is assumed to be much thicker than skin depth
        Rout = rho / (2*pi*delta*self.b)
        return Rin + Rout

    @property
    def L(self) -> NumberLike:
        """
        Distributed inductance L, in H/m

        Returns
        -------
        L : number, or array-like
            distributed inductance, in  H/m

        """
        return mu_0/(2.*pi)*log(self.b/self.a)

    @property
    def C(self) -> NumberLike:
        """
        Distributed capacitance C, in F/m

        Returns
        -------
        C : number, or array-like
            distributed capacitance, in F/m

        """
        return 2.*pi*self.epsilon_prime/log(self.b/self.a)

    @property
    def G(self) -> NumberLike:
        """
        Distributed conductance G, in S/m

        Returns
        -------
        G : number, or array-like
            distributed conductance, in S/m

        """
        f =  self.frequency.f
        return f*self.epsilon_second/log(self.b/self.a)

    def __str__(self):
        f=self.frequency
        f_s = f'Coaxial Media.  {f.f_scaled[0]}-{f.f_scaled[-1]}{f.unit}, '\
                f'{f.npoints} points.'
        z0 = self.z0
        z0_s = f'z0 = ({z0[0].real:.1f}, {z0[0].imag:.1f}j)-'\
                f'({z0[-1].real:.1f}, {z0[-1].imag:.1f}j) Ohm'
        if self.z0_port is not None:
            z0_port = self.z0_port
            z0_port_s = \
                f'z0_port = ({z0_port[0].real:.1f}, {z0_port[0].imag:.1f}j)-'\
                f'({z0_port[-1].real:.1f}, {z0_port[-1].imag:.1f}j) Ohm\n'
        else:
            z0_port_s = 'z0_port is not defined.'
        try:
            output =  \
                f'Coaxial Media.  {f_s}\n'\
                f'Dint = {self.Dint * 1e3:.2f} mm,'\
                f' Dout = {self.Dout * 1e3:.2f} mm\n' \
                f'{z0_s}\n{z0_port_s}'
        except(TypeError):
            output =  \
                f'Coaxial Media.  {f_s}\n'\
                f'Dint = {self.Dint[0] * 1e3:.2f} mm,'\
                f' Dout = {self.Dout[0] * 1e3:.2f} mm\n' \
                f'{z0_s}\n{z0_port_s}'
        return output



import os
import unittest

import numpy as np

import skrf as rf


class CitiTestCase(unittest.TestCase):
    """
    Test the IO of CITI files.
    """
    def setUp(self):
        """
        Sets up the test directory
        """
        self.test_dir = os.path.dirname(os.path.abspath(__file__))+'/MDIF_CITI_MDL/'

        # constructor from filename
        self.oneport_example1 = rf.Citi(self.test_dir + 'test_1p_citi.cti')
        self.oneport_example2 = rf.Citi(self.test_dir + 'test_1p_citi_2_ri.cti')
        self.twoport_example1 = rf.Citi(self.test_dir + 'test_2p_citi.cti')
        self.twoport_example2 = rf.Citi(self.test_dir + 'test_2p_citi_2.cti')
        self.twoport_example3 = rf.Citi(self.test_dir + 'test_2p_citi_2params.cti')
        self.twoport_example4 = rf.Citi(self.test_dir + 'test_2p_citi_2params_db.cti')
        self.twoport_example5 = rf.Citi(self.test_dir + 'test_2p_citi_3_ri.cti')
        self.fourport_example1 = rf.Citi(self.test_dir + 'test_4p_citi.cti')

        self.examples = [self.oneport_example1, self.oneport_example2,
                          self.twoport_example1, self.twoport_example2,
                          self.twoport_example3, self.twoport_example4,
                          self.twoport_example5, self.fourport_example1]

        # constructor from file-object
        file = open(self.test_dir + 'test_1p_citi.cti')
        self.oneport_example1_from_fo = rf.Citi(file)

    def test_to_networks(self):
        """ Test if CITI data are correctly converted into Networks """
        for ex in self.examples:
            for ntwk in ex.networks:
                self.assertIsInstance(ntwk, rf.Network)

    def test_to_networkset(self):
        """ Test if CITI data are correctly converted into NetworkSet """
        for example in self.examples:
            self.assertIsInstance(example.to_networkset(), rf.NetworkSet)

    def test_params(self):
        """ Test if the params are correctly parsed from the CITI files """
        self.assertEqual(self.oneport_example1.params, ['Cm'])
        self.assertEqual(self.twoport_example1.params, ['Cm'])
        self.assertEqual(self.fourport_example1.params, ['Cm'])

    def test_only_freq_in_var(self):
        """ File without any VAR except for freq should return non empty NetworkSet. """
        file = self.test_dir + 'test_2p_only_freq_VAR.cti'
        cti = rf.Citi(file)
        ns = cti.to_networkset()
        self.assertTrue(ns)  # not empty
        self.assertEqual(len(ns), 1)

    def test_values_1p_1(self):
        """ Test if the values are correctly parsed from the CITI files """
        deg = np.array([
         -0.0178919999, -0.0180179999, -0.0181439998, -0.0182699998,
         -0.0183959998, -0.0185219998, -0.0186479998, -0.0187739998,
         -0.0188999998, -0.0204479998, -0.0205919998, -0.0207359998,
         -0.0208799998, -0.0210239998, -0.0211679998, -0.0213119998,
         -0.0214559997, -0.0215999997, -0.0230039997, -0.0231659997,
         -0.0233279997, -0.0234899997, -0.0236519997, -0.0238139997,
         -0.0239759997, -0.0241379996, -0.0242999996, -0.0255599996,
         -0.0257399996, -0.0259199996, -0.0260999995, -0.0262799995,
         -0.0264599995, -0.0266399995, -0.0268199995, -0.0269999995])
        f = np.array([
                710000000, 715000000, 720000000, 725000000,
                730000000, 735000000, 740000000, 745000000,
                750000000,
        ])
        mag = np.ones(len(deg))
        s1p_1 = rf.magdeg_2_reim(mag, deg)
        ns_1p_1 = self.oneport_example1.to_networkset()
        self.assertEqual(len(ns_1p_1), 4)
        self.assertEqual(len(ns_1p_1[0]), 9)
        np.testing.assert_array_equal(ns_1p_1[0].f, f)

        np.testing.assert_array_almost_equal(ns_1p_1[0].s.squeeze(), s1p_1[:9])

    def test_values_1p_2(self):
        """ Test if the values are correctly parsed from the CITI files """
        s1p_2 = np.array([
            -1.31189E-3 - 1.47980E-3 * 1j,
            -3.67867E-3 - 0.67782E-3 * 1j,
            -3.43990E-3 + 0.58746E-3 * 1j,
            -2.70664E-4 - 9.76175E-4 * 1j,
            +0.65892E-4 - 9.61571E-4 * 1j])
        ns_1p_2 = self.oneport_example2.to_networkset()
        self.assertEqual(len(ns_1p_2), 1)
        self.assertEqual(len(ns_1p_2[0]), len(s1p_2))
        np.testing.assert_array_equal(ns_1p_2[0].f, np.array([1., 2., 3., 4., 5.]))
        np.testing.assert_array_almost_equal(ns_1p_2[0].s.squeeze(), s1p_2)

    def test_values_2p_1(self):
        """ Test if the values are correctly parsed from the CITI files """
        mag = np.array([[[0.999999951, 0.000312274295], [0.000312274295, 0.999999951]]])
        deg = np.array([[[-0.0178919994, 89.982108], [89.982108, -0.0178919994]]])
        s_2p_1 = rf.magdeg_2_reim(mag, deg)
        f = np.array([
                710000000, 715000000, 720000000, 725000000,
                730000000, 735000000, 740000000, 745000000,
                750000000,
        ])
        ns_2p_1 = self.twoport_example1.to_networkset()
        self.assertEqual(len(ns_2p_1), 4)
        self.assertEqual(len(ns_2p_1[0]), 9)
        np.testing.assert_array_equal(ns_2p_1[0].f, f)
        np.testing.assert_array_almost_equal(ns_2p_1[0].s_mag[0], mag.squeeze())
        np.testing.assert_array_almost_equal(ns_2p_1[0].s_deg[0], deg.squeeze())
        np.testing.assert_array_almost_equal(ns_2p_1[0].s[0], s_2p_1[0])

    def test_values_2p_2(self):
        """ Test if the values are correctly parsed from the CITI files """
        mag = np.array([
            [[0.1, 0.3],
             [0.5, 0.7]],
            [[0.2, 0.4],
             [0.6, 0.8]]
            ])
        deg = np.array([
            [[2, 4],
             [6, 8]],
            [[3, 5],
             [7, 9]]
            ])
        s2p_2 = rf.magdeg_2_reim(mag, deg)
        ns_2p_2 = self.twoport_example2.to_networkset()
        self.assertEqual(len(ns_2p_2), 1)
        self.assertEqual(len(ns_2p_2[0]), 2)
        np.testing.assert_array_equal(ns_2p_2[0].f, np.array([1e9, 2e9]))
        np.testing.assert_array_almost_equal(ns_2p_2[0].s.squeeze(), s2p_2)

    def test_values_4p(self):
        """ Test if the values are correctly parsed from the CITI files """
        ns_4p = self.fourport_example1.to_networkset()
        self.assertEqual(len(ns_4p), 3)
        self.assertEqual(len(ns_4p[0]), 51)
        self.assertEqual(ns_4p.coords['Cm'], [9e-16, 8e-16, 7e-16])

suite = unittest.TestLoader().loadTestsFromTestCase(CitiTestCase)
unittest.TextTestRunner(verbosity=2).run(suite)



r"""
.. module:: skrf.tlineFunctions
===============================================
tlineFunctions (:mod:`skrf.tlineFunctions`)
===============================================

This module provides functions related to transmission line theory.

Impedance and Reflection Coefficient
--------------------------------------
These functions relate basic transmission line quantities such as
characteristic impedance, input impedance, reflection coefficient, etc.
Each function has two names. One is a long-winded but readable name and
the other is a short-hand variable-like names. Below is a table relating
these two names with each other as well as common mathematical symbols.

====================  ======================  ================================
Symbol                Variable Name           Long Name
====================  ======================  ================================
:math:`Z_l`           z_l                     load_impedance
:math:`Z_{in}`        z_in                    input_impedance
:math:`\Gamma_0`      Gamma_0                 reflection_coefficient
:math:`\Gamma_{in}`   Gamma_in                reflection_coefficient_at_theta
:math:`\theta`        theta                   electrical_length
====================  ======================  ================================

There may be a bit of confusion about the difference between the load
impedance the input impedance. This is because the load impedance **is**
the input impedance at the load. An illustration may provide some
useful reference.

Below is a (bad) illustration of a section of uniform transmission line
of characteristic impedance :math:`Z_0`, and electrical length
:math:`\theta`. The line is terminated on the right with some
load impedance, :math:`Z_l`. The input impedance :math:`Z_{in}` and
input reflection coefficient :math:`\Gamma_{in}` are
looking in towards the load from the distance :math:`\theta` from the
load.

.. math::
        Z_0, \theta

        \text{o===============o=}[Z_l]

        \to\qquad\qquad\qquad\quad\qquad \qquad \to \qquad \quad

        Z_{in},\Gamma_{in}\qquad\qquad\qquad\qquad\quad Z_l,\Gamma_0

So, to clarify the confusion,

.. math::
        Z_{in}= Z_{l},\qquad\qquad
        \Gamma_{in}=\Gamma_l \text{ at }  \theta=0


Short names
+++++++++++++
.. autosummary::
        :toctree: generated/

        theta

        zl_2_Gamma0
        zl_2_zin
        zl_2_Gamma_in
        zl_2_swr
        zl_2_total_loss

        Gamma0_2_zl
        Gamma0_2_Gamma_in
        Gamma0_2_zin
        Gamma0_2_swr

Long-names
++++++++++++++
.. autosummary::
        :toctree: generated/

        electrical_length

        distance_2_electrical_length
        electrical_length_2_distance

        reflection_coefficient_at_theta
        reflection_coefficient_2_input_impedance
        reflection_coefficient_2_input_impedance_at_theta
        reflection_coefficient_2_propagation_constant

        input_impedance_at_theta
        load_impedance_2_reflection_coefficient
        load_impedance_2_reflection_coefficient_at_theta

        voltage_current_propagation



Distributed Circuit and Wave Quantities
----------------------------------------
.. autosummary::
        :toctree: generated/

        distributed_circuit_2_propagation_impedance
        propagation_impedance_2_distributed_circuit

Transmission Line Physics
---------------------------------
.. autosummary::
        :toctree: generated/

        skin_depth
        surface_resistivity
"""

import numpy as np
from numpy import array, exp, pi, real, sqrt
from scipy.constants import mu_0

from . import mathFunctions as mf
from .constants import INF, ONE, NumberLike


def skin_depth(f: NumberLike, rho: float, mu_r: float):
    r"""
    Skin depth for a material.

    The skin depth is calculated as:


    .. math::

        \delta = \sqrt{\frac{ \rho }{ \pi f \mu_r \mu_0 }}

    See www.microwaves101.com [#]_ or wikipedia [#]_ for more info.

    Parameters
    ----------
    f : number or array-like
        frequency, in Hz
    rho : number of array-like
        bulk resistivity of material, in ohm*m
    mu_r : number or array-like
        relative permeability of material

    Returns
    -------
    skin depth : number or array-like
        the skin depth, in meter

    References
    ----------
    .. [#] https://www.microwaves101.com/encyclopedias/skin-depth
    .. [#] http://en.wikipedia.org/wiki/Skin_effect

    See Also
    --------
    surface_resistivity

    """
    return sqrt(rho/(pi*f*mu_r*mu_0))


def surface_resistivity(f: NumberLike, rho: float, mu_r: float):
    r"""
    Surface resistivity.

    The surface resistivity is calculated as:


    .. math::

        \frac{ \rho }{ \delta }

    where :math:`\delta` is the skin depth from :func:`skin_depth`.

    See www.microwaves101.com [#]_ or wikipedia [#]_ for more info.

    Parameters
    ----------
    f : number or array-like
        frequency, in Hz
    rho : number or array-like
        bulk resistivity of material, in ohm*m
    mu_r : number or array-like
        relative permeability of material

    Returns
    -------
    surface resistivity : number of array-like
        Surface resistivity in ohms/square

    References
    ----------
    .. [#] https://www.microwaves101.com/encyclopedias/sheet-resistance
    .. [#] https://en.wikipedia.org/wiki/Sheet_resistance

    See Also
    --------
    skin_depth
    """
    return rho/skin_depth(rho=rho, f=f, mu_r=mu_r)


def distributed_circuit_2_propagation_impedance(distributed_admittance: NumberLike,
        distributed_impedance: NumberLike):
    r"""
    Convert distributed circuit values to wave quantities.

    This converts complex distributed impedance and admittance to
    propagation constant and characteristic impedance. The relation is

    .. math::
        Z_0 = \sqrt{ \frac{Z^{'}}{Y^{'}}}
        \quad\quad
        \gamma = \sqrt{ Z^{'}  Y^{'}}

    Parameters
    ----------
    distributed_admittance : number, array-like
        distributed admittance
    distributed_impedance :  number, array-like
        distributed impedance

    Returns
    -------
    propagation_constant : number, array-like
        distributed impedance
    characteristic_impedance : number, array-like
        distributed impedance

    See Also
    --------
        propagation_impedance_2_distributed_circuit : opposite conversion
    """
    propagation_constant = \
            sqrt(distributed_impedance*distributed_admittance)
    characteristic_impedance = \
            sqrt(distributed_impedance/distributed_admittance)
    return (propagation_constant, characteristic_impedance)


def propagation_impedance_2_distributed_circuit(propagation_constant: NumberLike,
        characteristic_impedance: NumberLike):
    r"""
    Convert wave quantities to distributed circuit values.

    Convert complex propagation constant and characteristic impedance
    to distributed impedance and admittance. The relation is,

    .. math::
        Z^{'} = \gamma  Z_0 \quad\quad
        Y^{'} = \frac{\gamma}{Z_0}

    Parameters
    ----------
    propagation_constant : number, array-like
        distributed impedance
    characteristic_impedance : number, array-like
        distributed impedance

    Returns
    -------
    distributed_admittance : number, array-like
        distributed admittance
    distributed_impedance :  number, array-like
        distributed impedance


    See Also
    --------
        distributed_circuit_2_propagation_impedance : opposite conversion
    """
    distributed_admittance = propagation_constant/characteristic_impedance
    distributed_impedance = propagation_constant*characteristic_impedance
    return (distributed_admittance, distributed_impedance)


def electrical_length(gamma: NumberLike, f: NumberLike, d: NumberLike, deg: bool = False):
    r"""
    Electrical length of a section of transmission line.

    .. math::
        \theta = \gamma(f) \cdot d

    Parameters
    ----------
    gamma : number, array-like or function
        propagation constant. See Notes.
        If passed as a function, takes frequency in Hz as a sole argument.
    f : number or array-like
        frequency at which to calculate
    d : number or array-like
        length of line, in meters
    deg : Boolean
        return in degrees or not.

    Returns
    -------
    theta : number or array-like
        electrical length in radians or degrees, depending on  value of deg.

    See Also
    --------
        electrical_length_2_distance : opposite conversion

    Note
    ----
    The convention has been chosen that forward propagation is
    represented by the positive imaginary part of the value returned by
    the gamma function.
    """
    # if gamma is not a function, create a dummy function which return gamma
    if not callable(gamma):
        _gamma = gamma
        def gamma(f0): return _gamma

    # typecast to a 1D array
    f = array(f, dtype=float).reshape(-1)
    d = array(d, dtype=float).reshape(-1)

    if not deg:
        return  gamma(f)*d
    else:
        return  mf.radian_2_degree(gamma(f)*d )


def electrical_length_2_distance(theta: NumberLike, gamma: NumberLike, f0: NumberLike, deg: bool = True):
    r"""
    Convert electrical length to a physical distance.

    .. math::
        d = \frac{\theta}{\gamma(f_0)}

    Parameters
    ----------
    theta : number or array-like
        electrical length. units depend on `deg` option
    gamma : number, array-like or function
        propagation constant. See Notes.
        If passed as a function, takes frequency in Hz as a sole argument.
    f0 : number or array-like
        frequency at which to calculate gamma
    deg : Boolean
        return in degrees or not.

    Returns
    -------
    d : number or array-like (real)
        physical distance in m

    Note
    ----
    The convention has been chosen that forward propagation is
    represented by the positive imaginary part of the value returned by
    the gamma function.

    See Also
    --------
        distance_2_electrical_length: opposite conversion
    """
    # if gamma is not a function, create a dummy function which return gamma
    if not callable(gamma):
        _gamma = gamma
        def gamma(f0): return _gamma

    if deg:
        theta = mf.degree_2_radian(theta)
    return real(theta / gamma(f0))


def load_impedance_2_reflection_coefficient(z0: NumberLike, zl: NumberLike):
    r"""
    Reflection coefficient from a load impedance.

    Return the reflection coefficient for a given load impedance, and
    characteristic impedance.

    For a transmission line of characteristic impedance :math:`Z_0`
    terminated with load impedance :math:`Z_l`, the complex reflection
    coefficient is given by,

    .. math::
        \Gamma = \frac {Z_l - Z_0}{Z_l + Z_0}

    Parameters
    ----------
    z0 : number or array-like
        characteristic impedance
    zl : number or array-like
        load impedance (aka input impedance)

    Returns
    -------
    gamma : number or array-like
        reflection coefficient

    See Also
    --------
        Gamma0_2_zl : reflection coefficient to load impedance

    Note
    ----
    Inputs are typecasted to 1D complex array.
    """
    # typecast to a complex 1D array. this makes everything easier
    z0 = array(z0, dtype=complex).reshape(-1)
    zl = array(zl, dtype=complex).reshape(-1)

    # handle singularity  by numerically representing inf as big number
    zl[(zl == np.inf)] = INF

    return ((zl - z0)/(zl + z0))


def reflection_coefficient_2_input_impedance(z0: NumberLike, Gamma: NumberLike):
    r"""
    Input impedance from a load reflection coefficient.

    Calculate the input impedance given a reflection coefficient and
    characteristic impedance.

    .. math::
        Z_0 \left(\frac {1 + \Gamma}{1-\Gamma} \right)

    Parameters
    ----------
    Gamma : number or array-like
        complex reflection coefficient
    z0 : number or array-like
        characteristic impedance

    Returns
    -------
    zin : number or array-like
        input impedance

    """
    # typecast to a complex 1D array. this makes everything easier
    Gamma = array(Gamma, dtype=complex).reshape(-1)
    z0 = array(z0, dtype=complex).reshape(-1)

    # handle singularity by numerically representing inf as close to 1
    Gamma[(Gamma == 1)] = ONE

    return z0*((1.0 + Gamma)/(1.0 - Gamma))


def reflection_coefficient_at_theta(Gamma0: NumberLike, theta: NumberLike):
    r"""
    Reflection coefficient at a given electrical length.

    .. math::
            \Gamma_{in} = \Gamma_0 e^{-2 \theta}

    Parameters
    ----------
    Gamma0 : number or array-like
        reflection coefficient at theta=0
    theta : number or array-like
        electrical length (may be complex)

    Returns
    -------
    Gamma_in : number or array-like
        input reflection coefficient

    """
    Gamma0 = array(Gamma0, dtype=complex).reshape(-1)
    theta = array(theta, dtype=complex).reshape(-1)
    return Gamma0 * exp(-2*theta)


def input_impedance_at_theta(z0: NumberLike, zl: NumberLike, theta: NumberLike):
    """
    Input impedance from load impedance at a given electrical length.

    Input impedance of load impedance zl at a given electrical length,
    given characteristic impedance z0.

    Parameters
    ----------
    z0 : number or array-like
        characteristic impedance
    zl : number or array-like
        load impedance
    theta : number or array-like
        electrical length of the line (may be complex)

    Returns
    -------
    zin : number or array-like
        input impedance at theta

    """
    Gamma0 = load_impedance_2_reflection_coefficient(z0=z0, zl=zl)
    Gamma_in = reflection_coefficient_at_theta(Gamma0=Gamma0, theta=theta)
    return reflection_coefficient_2_input_impedance(z0=z0, Gamma=Gamma_in)


def load_impedance_2_reflection_coefficient_at_theta(z0: NumberLike, zl: NumberLike, theta: NumberLike):
    """
    Reflection coefficient of load at a given electrical length.

    Reflection coefficient of load impedance zl at a given electrical length,
    given characteristic impedance z0.

    Parameters
    ----------
    z0 : number or array-like
        characteristic impedance.
    zl : number or array-like
        load impedance
    theta : number or array-like
        electrical length of the line (may be complex).

    Returns
    -------
    Gamma_in : number or array-like
        input reflection coefficient at theta

    """
    Gamma0 = load_impedance_2_reflection_coefficient(z0=z0, zl=zl)
    Gamma_in = reflection_coefficient_at_theta(Gamma0=Gamma0, theta=theta)
    return Gamma_in


def reflection_coefficient_2_input_impedance_at_theta(z0: NumberLike, Gamma0: NumberLike, theta: NumberLike):
    """
    Input impedance from load reflection coefficient at a given electrical length.

    Calculate the input impedance at electrical length theta, given a
    reflection coefficient and characteristic impedance of the medium.

    Parameters
    ----------
    z0 : number or array-like
        characteristic impedance.
    Gamma: number or array-like
        reflection coefficient
    theta: number or array-like
        electrical length of the line, (may be complex)

    Returns
    -------
    zin: number or array-like
        input impedance at theta

    """
    Gamma_in = reflection_coefficient_at_theta(Gamma0=Gamma0, theta=theta)
    zin = reflection_coefficient_2_input_impedance(z0=z0, Gamma=Gamma_in)
    return zin


def reflection_coefficient_2_propagation_constant(Gamma_in: NumberLike, Gamma_l: NumberLike, d: NumberLike):
    r"""
    Propagation constant from line input and load reflection coefficients.

    Calculate the propagation constant of a line of length d, given the
    reflection coefficient and characteristic impedance of the medium.

    .. math::
        \Gamma_{in} = \Gamma_l e^{-2 j \gamma \cdot d}
        \to \gamma = -\frac{1}{2 d} \ln \left ( \frac{ \Gamma_{in} }{ \Gamma_l } \right )

    Parameters
    ----------
    Gamma_in : number or array-like
        input reflection coefficient
    Gamma_l :  number or array-like
        load reflection coefficient
    d : number or array-like
        length of line, in meters

    Returns
    -------
    gamma : number (complex) or array-like
        propagation constant (see notes)

    Note
    ----
    The convention has been chosen that forward propagation is
    represented by the positive imaginary part of gamma.

    """
    gamma = -1/(2*d) * np.log(Gamma_in/Gamma_l)
    # the imaginary part of gamma (=beta) cannot be negative with the given
    # definition of gamma. Thus one should take the first modulo positive value
    gamma.imag = gamma.imag % (pi/d)

    return gamma


def Gamma0_2_swr(Gamma0: NumberLike):
    r"""
    Standing Wave Ratio (SWR) for a given reflection coefficient.

    Standing Wave Ratio value is defined by:

    .. math::
        VSWR = \frac{1 + |\Gamma_0|}{1 - |\Gamma_0|}

    Parameters
    ----------
    Gamma0 : number or array-like
        Reflection coefficient

    Returns
    -------
    swr : number or array-like
        Standing Wave Ratio.

    """
    return (1 + np.abs(Gamma0)) / (1 - np.abs(Gamma0))


def zl_2_swr(z0: NumberLike, zl: NumberLike):
    r"""
    Standing Wave Ratio (SWR) for a given load impedance.

    Standing Wave Ratio value is defined by:

    .. math::
        VSWR = \frac{1 + |\Gamma|}{1 - |\Gamma|}

    where

    .. math::
        \Gamma = \frac{Z_L - Z_0}{Z_L + Z_0}

    Parameters
    ----------
    z0 : number or array-like
        line characteristic impedance [Ohm]
    zl : number or array-like
        load impedance [Ohm]

    Returns
    -------
    swr : number or array-like
        Standing Wave Ratio.

    """
    Gamma0 = load_impedance_2_reflection_coefficient(z0, zl)
    return Gamma0_2_swr(Gamma0)


def voltage_current_propagation(v1: NumberLike, i1: NumberLike, z0: NumberLike, theta: NumberLike):
    """
    Voltages and currents calculated on electrical length theta of a transmission line.

    Give voltage v2 and current i1 at theta, given voltage v1
    and current i1 at theta=0 and given characteristic parameters gamma and z0.

    ::

        i1                          i2
        ○-->---------------------->--○

        v1         gamma,z0         v2

        ○----------------------------○

        <------------ d ------------->

        theta=0                   theta

    Uses (inverse) ABCD parameters of a transmission line.

    Parameters
    ----------
    v1 : array-like (nfreqs,)
        total voltage at z=0
    i1 : array-like (nfreqs,)
        total current at z=0, directed toward the transmission line
    z0: array-like (nfreqs,)
        characteristic impedance
    theta : number or array-like (nfreq, ntheta)
        electrical length of the line (may be complex).

    Return
    ------
    v2 : array-like (nfreqs, ntheta)
        total voltage at z=d
    i2 : array-like (nfreqs, ndtheta
        total current at z=d, directed outward the transmission line
    """
    # outer product by broadcasting of the electrical length
    # theta = gamma[:, np.newaxis] * d  # (nbfreqs x nbd)
    # ABCD parameters of a transmission line (gamma, z0)
    A = np.cosh(theta)
    B = z0*np.sinh(theta)
    C = np.sinh(theta)/z0
    D = np.cosh(theta)
    # transpose and de-transpose operations are necessary
    # for linalg.inv to inverse square matrices
    ABCD = np.array([[A, B],[C, D]]).transpose()
    inv_ABCD = np.linalg.inv(ABCD).transpose()

    v2 = inv_ABCD[0,0] * v1 + inv_ABCD[0,1] * i1
    i2 = inv_ABCD[1,0] * v1 + inv_ABCD[1,1] * i1
    return v2, i2


def zl_2_total_loss(z0: NumberLike, zl: NumberLike, theta: NumberLike):
    r"""
    Total loss of a terminated transmission line (in natural unit).

    The total loss expressed in terms of the load impedance is [#]_ :

    .. math::
        TL = \frac{R_{in}}{R_L} \left| \cosh \theta  + \frac{Z_L}{Z_0} \sinh\theta \right|^2

    Parameters
    ----------
    z0 : number or array-like
        characteristic impedance.
    zl : number or array-like
        load impedance
    theta : number or array-like
        electrical length of the line (may be complex).

    Returns
    -------
    total_loss: number or array-like
        total loss in natural unit

    References
    ----------
    .. [#] Steve Stearns (K6OIK), Transmission Line Power Paradox and Its Resolution.
        ARRL PacificonAntenna Seminar, Santa Clara, CA, October 10-12, 2014.
        https://www.fars.k6ya.org/docs/K6OIK-A_Transmission_Line_Power_Paradox_and_Its_Resolution.pdf

    """
    Rin = np.real(zl_2_zin(z0, zl, theta))
    total_loss = Rin/np.real(zl)*np.abs(np.cosh(theta) + zl/z0*np.sinh(theta))**2
    return total_loss


# short hand convenience.
# admittedly these follow no logical naming scheme, but they closely
# correspond to common symbolic conventions, and are convenient
theta = electrical_length
distance_2_electrical_length = electrical_length

zl_2_Gamma0 = load_impedance_2_reflection_coefficient
Gamma0_2_zl = reflection_coefficient_2_input_impedance

zl_2_zin = input_impedance_at_theta
zl_2_Gamma_in = load_impedance_2_reflection_coefficient_at_theta

Gamma0_2_Gamma_in = reflection_coefficient_at_theta
Gamma0_2_zin = reflection_coefficient_2_input_impedance_at_theta
